from collections import OrderedDict

from django.contrib.auth.models import Permission

from openwisp_controller.migrations import create_default_permissions, get_swapped_model

# Use a pre-defined UUID so the template can be upgraded via migration scripts if needed
TEMPLATE_MONITORING_UUID = '00000000-defa-defa-defa-000000000000'
TEMPLATE_OPENWISP_MONITORING_01 = OrderedDict(
    {
        "path": "/usr/sbin/openwisp-monitoring",
        "mode": "0744",
        "contents": "uuid=$(uci get openwisp.http.uuid)\nkey=$(uci get openwisp.http.key)\nbase_url=$(uci get openwisp.http.url)\nverify_ssl=$(uci get openwisp.http.verify_ssl)\nincluded_interfaces=$(uci get openwisp.monitoring.included_interfaces)\nurl=\"$base_url/api/v1/monitoring/device/$uuid/?key=$key\"\ndata=$(/usr/sbin/netjson-monitoring \"$included_interfaces\")\nif [ \"$verify_ssl\" = 0 ]; then\n    curl_command='curl -k'\nelse\n    curl_command='curl'\nfi\n# send data via POST\n$curl_command -H \"Content-Type: application/json\" \\\n              -X POST \\\n              -d \"$data\" \\\n              -v $url\n",  # noqa
    }
)
TEMPLATE_OPENWISP_MONITORING_02 = OrderedDict(
    {
        "path": "/usr/sbin/legacy-openwisp-monitoring",
        "mode": "0744",
        "contents": "#!/bin/sh\n\n# Exit if openwisp-monitoring package is installed\nopenwisp_monitoring_installed=$(opkg list-installed | grep openwisp-monitoring -c)\nif [ \"$openwisp_monitoring_installed\" == \"1\" ]; then\n    exit 0\nfi\nuuid=$(uci get openwisp.http.uuid)\nkey=$(uci get openwisp.http.key)\nbase_url=$(uci get openwisp.http.url)\nverify_ssl=$(uci get openwisp.http.verify_ssl)\nincluded_interfaces=$(uci get openwisp.monitoring.included_interfaces)\nurl=\"$base_url/api/v1/monitoring/device/$uuid/?key=$key\"\ndata=$(/usr/sbin/legacy-netjson-monitoring \"$included_interfaces\")\nif [ \"$verify_ssl\" = 0 ]; then\n    curl_command='curl -k'\nelse\n    curl_command='curl'\nfi\n# send data via POST\n$curl_command -H \"Content-Type: application/json\" \\\n              -v -d \"$data\" $url\n",  # noqa
    }
)
TEMPLATE_NETJSON_MONITORING_01 = OrderedDict(
    {
        "path": "/usr/sbin/netjson-monitoring",
        "mode": "0744",
        "contents": "#!/usr/bin/env lua\n-- retrieve monitoring information\n-- and return it as NetJSON Output\nio = require('io')\nubus_lib = require('ubus')\ncjson = require('cjson')\nnixio = require('nixio')\nuci = require('uci')\nuci_cursor = uci.cursor()\n\n-- split function\nfunction split(str, pat)\n    local t = {}\n    local fpat = \"(.-)\" .. pat\n    local last_end = 1\n    local s, e, cap = str:find(fpat, 1)\n    while s do\n        if s ~= 1 or cap ~= \"\" then\n            table.insert(t, cap)\n        end\n        last_end = e + 1\n        s, e, cap = str:find(fpat, last_end)\n    end\n    if last_end <= #str then\n        cap = str:sub(last_end)\n        table.insert(t, cap)\n    end\n    return t\nend\n\nlocal function has_value(tab, val)\n    for index, value in ipairs(tab) do\n        if value == val then\n            return true\n        end\n    end\n    return false\nend\n\nlocal function starts_with(str, start)\n    return str:sub(1, #start) == start\nend\n\n-- parse /proc/net/arp\nfunction parse_arp()\n    arp_info = {}\n    for line in io.lines('/proc/net/arp 2> /dev/null') do\n        if line:sub(1, 10) ~= 'IP address' then\n            ip, hw, flags, mac, mask, dev = line:match(\"(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)\")\n            table.insert(arp_info, {\n                ip = ip,\n                mac = mac,\n                interface = dev,\n                state = ''\n            })\n        end\n    end\n    return arp_info\nend\n\nfunction get_ip_neigh_json()\n    arp_info = {}\n    output = io.popen('ip -json neigh 2> /dev/null'):read()\n    if output ~= nil and pcall(function () json_output = cjson.decode(output) end) then\n        json_output = cjson.decode(output)\n        for _, arp_entry in pairs(json_output) do\n            table.insert(arp_info, {\n                ip = arp_entry[\"dst\"],\n                mac = arp_entry[\"lladdr\"],\n                interface = arp_entry[\"dev\"],\n                state = arp_entry[\"state\"][1]\n            })\n        end\n    end\n    return arp_info\nend\n\nfunction get_ip_neigh()\n    arp_info = {}\n    output = io.popen('ip neigh 2> /dev/null')\n    for line in output:lines() do\n        ip, dev, mac, state = line:match(\"(%S+)%s+dev%s+(%S+)%s+lladdr%s+(%S+).*%s(%S+)\")\n        if mac ~= nil then\n            table.insert(arp_info, {\n                ip = ip,\n                mac = mac,\n                interface = dev,\n                state = state\n            })\n        end\n    end\n    return arp_info\nend\n\nfunction get_neighbors()\n    arp_table = get_ip_neigh_json()\n    if next(arp_table) == nil then\n        arp_table = get_ip_neigh()\n    end\n    if next(arp_table) == nil then\n        arp_table = parse_arp()\n    end\n    return arp_table\nend\n\nfunction parse_dhcp_lease_file(path, leases)\n    local f = io.open(path, 'r')\n    if not f then\n        return leases\n    end\n\n    for line in f:lines() do\n        local expiry, mac, ip, name, id = line:match('(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)')\n        table.insert(leases, {\n            expiry = tonumber(expiry),\n            mac = mac,\n            ip = ip,\n            client_name = name,\n            client_id = id\n        })\n    end\n\n    return leases\nend\n\nfunction get_dhcp_leases()\n    local dhcp_configs = uci_cursor:get_all('dhcp')\n    local leases = {}\n\n    if not dhcp_configs or not next(dhcp_configs) then\n        return nil\n    end\n\n    for name, config in pairs(dhcp_configs) do\n        if config and config['.type'] == 'dnsmasq' and config.leasefile then\n            leases = parse_dhcp_lease_file(config.leasefile, leases)\n        end\n    end\n    return leases\nend\n\nfunction is_table_empty(table_)\n    return not table_ or next(table_) == nil\nend\n\nfunction parse_hostapd_clients(clients)\n  local data = {}\n  for mac, properties in pairs(clients) do\n      properties.mac = mac\n      table.insert(data, properties)\n  end\n  return data\nend\n\nfunction parse_iwinfo_clients(clients)\n  local data = {}\n  for i, p in pairs(clients) do\n      client = {}\n      client.ht = p.rx.ht\n      client.mac = p.mac\n      client.authorized = p.authorized\n      client.vht = p.rx.vht\n      client.wmm = p.wme\n      client.mfp = p.mfp\n      client.auth = p.authenticated\n      client.signal = p.signal\n      client.noise = p.noise\n      table.insert(data, client)\n  end\n  return data\nend\n\n-- takes ubus wireless.status clients output and converts it to NetJSON\nfunction netjson_clients(clients, is_mesh)\n    return (is_mesh and parse_iwinfo_clients(clients) or parse_hostapd_clients(clients))\nend\n\nubus = ubus_lib.connect()\nif not ubus then\n    error('Failed to connect to ubusd')\nend\n\n-- helpers\niwinfo_modes = {\n    ['Master'] = 'access_point',\n    ['Client'] = 'station',\n    ['Mesh Point'] = '802.11s',\n    ['Ad-Hoc'] = 'adhoc'\n}\n\n-- collect system info\nsystem_info = ubus:call('system', 'info', {})\nboard = ubus:call('system', 'board', {})\nloadavg_output = io.popen('cat /proc/loadavg'):read()\nloadavg_output = split(loadavg_output, ' ')\nload_average = {tonumber(loadavg_output[1]), tonumber(loadavg_output[2]), tonumber(loadavg_output[3])}\n\nfunction parse_disk_usage()\n    file = io.popen('df')\n    disk_usage_info = {}\n    for line in file:lines() do\n        if line:sub(1, 10) ~= 'Filesystem' then\n            filesystem, size, used, available, percent, location =\n                line:match('(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)')\n            if filesystem ~= 'tmpfs' and not string.match(filesystem, 'overlayfs') then\n                percent = percent:gsub('%W', '')\n                -- available, size and used are in KiB\n                table.insert(disk_usage_info, {\n                    filesystem = filesystem,\n                    available_bytes = tonumber(available) * 1024,\n                    size_bytes = tonumber(size) * 1024,\n                    used_bytes = tonumber(used) * 1024,\n                    used_percent = tonumber(percent),\n                    mount_point = location\n                })\n            end\n        end\n    end\n    file:close()\n    return disk_usage_info\nend\n\nfunction get_cpus()\n    processors = io.popen('cat /proc/cpuinfo | grep -c processor')\n    cpus = tonumber(processors:read('*a'))\n    processors:close()\n    return cpus\nend\n\nfunction get_vpn_interfaces()\n    -- only openvpn supported for now\n    local items = uci_cursor:get_all('openvpn')\n    local vpn_interfaces = {}\n\n    if is_table_empty(items) then\n        return {}\n    end\n\n    for name, config in pairs(items) do\n        if config and config.dev then\n            vpn_interfaces[config.dev] = true\n        end\n    end\n    return vpn_interfaces\nend\n\n-- init netjson data structure\nnetjson = {\n    type = 'DeviceMonitoring',\n    general = {\n        hostname = board.hostname,\n        local_time = system_info.localtime,\n        uptime = system_info.uptime\n    },\n    resources = {\n        load = load_average,\n        memory = system_info.memory,\n        swap = system_info.swap,\n        cpus = get_cpus(),\n        disk = parse_disk_usage()\n    }\n}\n\ndhcp_leases = get_dhcp_leases()\nif not is_table_empty(dhcp_leases) then\n    netjson.dhcp_leases = dhcp_leases\nend\n\nneighbors = get_neighbors()\nif not is_table_empty(neighbors) then\n    netjson.neighbors = neighbors\nend\n\n-- determine the interfaces to monitor\ntraffic_monitored = arg[1]\ninclude_stats = {}\nif traffic_monitored then\n    traffic_monitored = split(traffic_monitored, ' ')\n    for i, name in pairs(traffic_monitored) do\n        include_stats[name] = true\n    end\nend\n\nfunction is_excluded(name)\n    return name == 'lo'\nend\n\nfunction find_default_gateway(routes)\n    for i = 1, #routes do\n        if routes[i].target == '0.0.0.0' then\n            return routes[i].nexthop\n        end\n    end\n    return nil\nend\n\n-- collect device data\nnetwork_status = ubus:call('network.device', 'status', {})\nwireless_status = ubus:call('network.wireless', 'status', {})\ninterface_data = ubus:call('network.interface', 'dump', {})\nnixio_data = nixio.getifaddrs()\nvpn_interfaces = get_vpn_interfaces()\nwireless_interfaces = {}\ninterfaces = {}\ndns_servers = {}\ndns_search = {}\n\nfunction new_address_array(address, interface, family)\n    proto = interface['proto']\n    if proto == 'dhcpv6' then\n        proto = 'dhcp'\n    end\n    new_address = {\n        address = address['address'],\n        mask = address['mask'],\n        proto = proto,\n        family = family,\n        gateway = find_default_gateway(interface.route)\n    }\n    return new_address\nend\n\nfunction get_interface_info(name, netjson_interface)\n    info = {\n        dns_search = nil,\n        dns_servers = nil\n    }\n    for _, interface in pairs(interface_data['interface']) do\n        if interface['l3_device'] == name then\n            if next(interface['dns-search']) then\n                info.dns_search = interface['dns-search']\n            end\n            if next(interface['dns-server']) then\n                info.dns_servers = interface['dns-server']\n            end\n            if netjson_interface.type == 'bridge' then\n                info.stp = uci_cursor.get('network', interface['interface'], 'stp') == '1'\n            end\n        end\n    end\n    return info\nend\n\nfunction array_concat(source, destination)\n    table.foreach(source, function(key, value)\n        table.insert(destination, value)\n    end)\nend\n\nfunction dict_merge(source, destination)\n    table.foreach(source, function(key, value)\n        destination[key] = value\n    end)\nend\n\n-- collect interface addresses\nfunction get_addresses(name)\n    addresses = {}\n    interface_list = interface_data['interface']\n    addresses_list = {}\n    for _, interface in pairs(interface_list) do\n        if interface['l3_device'] == name then\n            proto = interface['proto']\n            if proto == 'dhcpv6' then\n                proto = 'dhcp'\n            end\n            for _, address in pairs(interface['ipv4-address']) do\n                table.insert(addresses_list, address['address'])\n                new_address = new_address_array(address, interface, 'ipv4')\n                table.insert(addresses, new_address)\n            end\n            for _, address in pairs(interface['ipv6-address']) do\n                table.insert(addresses_list, address['address'])\n                new_address = new_address_array(address, interface, 'ipv6')\n                table.insert(addresses, new_address)\n            end\n        end\n    end\n    for i = 1, #nixio_data do\n        if nixio_data[i].name == name then\n            if not is_excluded(name) then\n                family = nixio_data[i].family\n                addr = nixio_data[i].addr\n                if family == 'inet' then\n                    family = 'ipv4'\n                    -- Since we don't already know this from the dump, we can\n                    -- consider this dynamically assigned, this is the case for\n                    -- example for OpenVPN interfaces, which get their address\n                    -- from the DHCP server embedded in OpenVPN\n                    proto = 'dhcp'\n                elseif family == 'inet6' then\n                    family = 'ipv6'\n                    if starts_with(addr, 'fe80') then\n                        proto = 'static'\n                    else\n                        ula = uci_cursor.get('network', 'globals', 'ula_prefix')\n                        ula_prefix = split(ula, '::')[1]\n                        if starts_with(addr, ula_prefix) then\n                            proto = 'static'\n                        else\n                            proto = 'dhcp'\n                        end\n                    end\n                end\n                if family == 'ipv4' or family == 'ipv6' then\n                    if not has_value(addresses_list, addr) then\n                        table.insert(addresses, {\n                            address = addr,\n                            mask = nixio_data[i].prefix,\n                            proto = proto,\n                            family = family\n                        })\n                    end\n                end\n            end\n        end\n    end\n    return addresses\nend\n\n-- collect relevant wireless interface stats\n-- (traffic and connected clients)\nfor radio_name, radio in pairs(wireless_status) do\n    for i, interface in ipairs(radio.interfaces) do\n        name = interface.ifname\n        local is_mesh = false\n        if name and not is_excluded(name) then\n            iwinfo = ubus:call('iwinfo', 'info', {\n                device = name\n            })\n            netjson_interface = {\n                name = name,\n                type = 'wireless',\n                wireless = {\n                    ssid = iwinfo.ssid,\n                    mode = iwinfo_modes[iwinfo.mode] or iwinfo.mode,\n                    channel = iwinfo.channel,\n                    frequency = iwinfo.frequency,\n                    tx_power = iwinfo.txpower,\n                    signal = iwinfo.signal,\n                    noise = iwinfo.noise,\n                    country = iwinfo.country\n                }\n            }\n            if iwinfo.mode == 'Ad-Hoc' or iwinfo.mode == 'Mesh Point' then\n              clients = ubus:call('iwinfo', 'assoclist', {\n                device = name\n              }).results\n              is_mesh = true\n            else\n              clients = ubus:call('hostapd.' .. name, 'get_clients', {}).clients\n            end\n            if clients and next(clients) ~= nil then\n                netjson_interface.wireless.clients = netjson_clients(clients, is_mesh)\n            end\n            wireless_interfaces[name] = netjson_interface\n        end\n    end\nend\n\nfunction needs_inversion(interface)\n    return interface.type == 'wireless' and interface.wireless.mode == 'access_point'\nend\n\nfunction invert_rx_tx(interface)\n    for k, v in pairs(interface) do\n        if string.sub(k, 0, 3) == \"rx_\" then\n            local tx_key = \"tx_\" .. string.sub(k, 4)\n            local tx_val = interface[tx_key]\n            interface[tx_key] = v\n            interface[k] = tx_val\n        end\n    end\n    return interface\nend\n\n-- collect interface stats\nfor name, interface in pairs(network_status) do\n    -- only collect data from iterfaces which have not been excluded\n    if not is_excluded(name) then\n        netjson_interface = {\n            name = name,\n            type = string.lower(interface.type),\n            up = interface.up,\n            mac = interface.macaddr,\n            txqueuelen = interface.txqueuelen,\n            mtu = interface.mtu,\n            speed = interface.speed,\n            bridge_members = interface['bridge-members'],\n            multicast = interface.multicast\n        }\n        if wireless_interfaces[name] then\n            dict_merge(wireless_interfaces[name], netjson_interface)\n            interface.type = netjson_interface.type\n        end\n        if interface.type == 'Network device' then\n            link_supported = interface['link-supported']\n            if link_supported and next(link_supported) then\n                netjson_interface.type = 'ethernet'\n                netjson_interface.link_supported = link_supported\n            elseif vpn_interfaces[name] then\n                netjson_interface.type = 'virtual'\n            else\n                netjson_interface.type = 'other'\n            end\n        end\n        if include_stats[name] then\n            if needs_inversion(netjson_interface) then\n                --- ensure wifi access point interfaces\n                --- show download and upload values from\n                --- the user's perspective and not from the router perspective\n                interface.statistics = invert_rx_tx(interface.statistics)\n            end\n            netjson_interface.statistics = interface.statistics\n        end\n        addresses = get_addresses(name)\n        if next(addresses) then\n            netjson_interface.addresses = addresses\n        end\n        info = get_interface_info(name, netjson_interface)\n        if info.stp ~= nil then\n            netjson_interface.stp = info.stp\n        end\n        table.insert(interfaces, netjson_interface)\n        -- DNS info is independent from interface\n        if info.dns_servers then\n            array_concat(info.dns_servers, dns_servers)\n        end\n        if info.dns_search then\n            array_concat(info.dns_search, dns_search)\n        end\n    end\nend\n\nif next(interfaces) ~= nil then\n    netjson.interfaces = interfaces\nend\nif next(dns_servers) ~= nil then\n    netjson.dns_servers = dns_servers\nend\nif next(dns_search) ~= nil then\n    netjson.dns_search = dns_search\nend\n\nprint(cjson.encode(netjson))\n",  # noqa
    }
)
TEMPLATE_NETJSON_MONITORING_02 = OrderedDict(
    {
        "path": "/usr/sbin/legacy-netjson-monitoring",
        "mode": "0744",
        "contents": "#!/usr/bin/env lua\n-- retrieve monitoring information\n-- and return it as NetJSON Output\nio = require('io')\nubus_lib = require('ubus')\ncjson = require('cjson')\nnixio = require('nixio')\nuci = require('uci')\nuci_cursor = uci.cursor()\n\n-- split function\nfunction split(str, pat)\n    local t = {}\n    local fpat = \"(.-)\" .. pat\n    local last_end = 1\n    local s, e, cap = str:find(fpat, 1)\n    while s do\n        if s ~= 1 or cap ~= \"\" then\n            table.insert(t, cap)\n        end\n        last_end = e + 1\n        s, e, cap = str:find(fpat, last_end)\n    end\n    if last_end <= #str then\n        cap = str:sub(last_end)\n        table.insert(t, cap)\n    end\n    return t\nend\n\nlocal function has_value(tab, val)\n    for index, value in ipairs(tab) do\n        if value == val then\n            return true\n        end\n    end\n    return false\nend\n\nlocal function starts_with(str, start)\n    return str:sub(1, #start) == start\nend\n\n-- parse /proc/net/arp\nfunction parse_arp()\n    arp_info = {}\n    for line in io.lines('/proc/net/arp 2> /dev/null') do\n        if line:sub(1, 10) ~= 'IP address' then\n            ip, hw, flags, mac, mask, dev = line:match(\"(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)\")\n            table.insert(arp_info, {\n                ip = ip,\n                mac = mac,\n                interface = dev,\n                state = ''\n            })\n        end\n    end\n    return arp_info\nend\n\nfunction get_ip_neigh_json()\n    arp_info = {}\n    output = io.popen('ip -json neigh 2> /dev/null'):read()\n    if output ~= nil and pcall(function () json_output = cjson.decode(output) end) then\n        json_output = cjson.decode(output)\n        for _, arp_entry in pairs(json_output) do\n            table.insert(arp_info, {\n                ip = arp_entry[\"dst\"],\n                mac = arp_entry[\"lladdr\"],\n                interface = arp_entry[\"dev\"],\n                state = arp_entry[\"state\"][1]\n            })\n        end\n    end\n    return arp_info\nend\n\nfunction get_ip_neigh()\n    arp_info = {}\n    output = io.popen('ip neigh 2> /dev/null')\n    for line in output:lines() do\n        ip, dev, mac, state = line:match(\"(%S+)%s+dev%s+(%S+)%s+lladdr%s+(%S+).*%s(%S+)\")\n        if mac ~= nil then\n            table.insert(arp_info, {\n                ip = ip,\n                mac = mac,\n                interface = dev,\n                state = state\n            })\n        end\n    end\n    return arp_info\nend\n\nfunction get_neighbors()\n    arp_table = get_ip_neigh_json()\n    if next(arp_table) == nil then\n        arp_table = get_ip_neigh()\n    end\n    if next(arp_table) == nil then\n        arp_table = parse_arp()\n    end\n    return arp_table\nend\n\nfunction parse_dhcp_lease_file(path, leases)\n    local f = io.open(path, 'r')\n    if not f then\n        return leases\n    end\n\n    for line in f:lines() do\n        local expiry, mac, ip, name, id = line:match('(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)')\n        table.insert(leases, {\n            expiry = tonumber(expiry),\n            mac = mac,\n            ip = ip,\n            client_name = name,\n            client_id = id\n        })\n    end\n\n    return leases\nend\n\nfunction get_dhcp_leases()\n    local dhcp_configs = uci_cursor:get_all('dhcp')\n    local leases = {}\n\n    if not dhcp_configs or not next(dhcp_configs) then\n        return nil\n    end\n\n    for name, config in pairs(dhcp_configs) do\n        if config and config['.type'] == 'dnsmasq' and config.leasefile then\n            leases = parse_dhcp_lease_file(config.leasefile, leases)\n        end\n    end\n    return leases\nend\n\nfunction is_table_empty(table_)\n    return not table_ or next(table_) == nil\nend\n\nfunction parse_hostapd_clients(clients)\n  local data = {}\n  for mac, properties in pairs(clients) do\n      properties.mac = mac\n      table.insert(data, properties)\n  end\n  return data\nend\n\nfunction parse_iwinfo_clients(clients)\n  local data = {}\n  for i, p in pairs(clients) do\n      client = {}\n      client.ht = p.rx.ht\n      client.mac = p.mac\n      client.authorized = p.authorized\n      client.vht = p.rx.vht\n      client.wmm = p.wme\n      client.mfp = p.mfp\n      client.auth = p.authenticated\n      client.signal = p.signal\n      client.noise = p.noise\n      table.insert(data, client)\n  end\n  return data\nend\n\n-- takes ubus wireless.status clients output and converts it to NetJSON\nfunction netjson_clients(clients, is_mesh)\n    return (is_mesh and parse_iwinfo_clients(clients) or parse_hostapd_clients(clients))\nend\n\nubus = ubus_lib.connect()\nif not ubus then\n    error('Failed to connect to ubusd')\nend\n\n-- helpers\niwinfo_modes = {\n    ['Master'] = 'access_point',\n    ['Client'] = 'station',\n    ['Mesh Point'] = '802.11s',\n    ['Ad-Hoc'] = 'adhoc'\n}\n\n-- collect system info\nsystem_info = ubus:call('system', 'info', {})\nboard = ubus:call('system', 'board', {})\nloadavg_output = io.popen('cat /proc/loadavg'):read()\nloadavg_output = split(loadavg_output, ' ')\nload_average = {tonumber(loadavg_output[1]), tonumber(loadavg_output[2]), tonumber(loadavg_output[3])}\n\nfunction parse_disk_usage()\n    file = io.popen('df')\n    disk_usage_info = {}\n    for line in file:lines() do\n        if line:sub(1, 10) ~= 'Filesystem' then\n            filesystem, size, used, available, percent, location =\n                line:match('(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)%s+(%S+)')\n            if filesystem ~= 'tmpfs' and not string.match(filesystem, 'overlayfs') then\n                percent = percent:gsub('%W', '')\n                -- available, size and used are in KiB\n                table.insert(disk_usage_info, {\n                    filesystem = filesystem,\n                    available_bytes = tonumber(available) * 1024,\n                    size_bytes = tonumber(size) * 1024,\n                    used_bytes = tonumber(used) * 1024,\n                    used_percent = tonumber(percent),\n                    mount_point = location\n                })\n            end\n        end\n    end\n    file:close()\n    return disk_usage_info\nend\n\nfunction get_cpus()\n    processors = io.popen('cat /proc/cpuinfo | grep -c processor')\n    cpus = tonumber(processors:read('*a'))\n    processors:close()\n    return cpus\nend\n\nfunction get_vpn_interfaces()\n    -- only openvpn supported for now\n    local items = uci_cursor:get_all('openvpn')\n    local vpn_interfaces = {}\n\n    if is_table_empty(items) then\n        return {}\n    end\n\n    for name, config in pairs(items) do\n        if config and config.dev then\n            vpn_interfaces[config.dev] = true\n        end\n    end\n    return vpn_interfaces\nend\n\n-- init netjson data structure\nnetjson = {\n    type = 'DeviceMonitoring',\n    general = {\n        hostname = board.hostname,\n        local_time = system_info.localtime,\n        uptime = system_info.uptime\n    },\n    resources = {\n        load = load_average,\n        memory = system_info.memory,\n        swap = system_info.swap,\n        cpus = get_cpus(),\n        disk = parse_disk_usage()\n    }\n}\n\ndhcp_leases = get_dhcp_leases()\nif not is_table_empty(dhcp_leases) then\n    netjson.dhcp_leases = dhcp_leases\nend\n\nneighbors = get_neighbors()\nif not is_table_empty(neighbors) then\n    netjson.neighbors = neighbors\nend\n\n-- determine the interfaces to monitor\ntraffic_monitored = arg[1]\ninclude_stats = {}\nif traffic_monitored and traffic_monitored ~= '*' then\n    traffic_monitored = split(traffic_monitored, ' ')\n    for i, name in pairs(traffic_monitored) do\n        include_stats[name] = true\n    end\nend\n\nfunction is_excluded(name)\n    return name == 'lo'\nend\n\nfunction find_default_gateway(routes)\n    for i = 1, #routes do\n        if routes[i].target == '0.0.0.0' then\n            return routes[i].nexthop\n        end\n    end\n    return nil\nend\n\n-- collect device data\nnetwork_status = ubus:call('network.device', 'status', {})\nwireless_status = ubus:call('network.wireless', 'status', {})\ninterface_data = ubus:call('network.interface', 'dump', {})\nnixio_data = nixio.getifaddrs()\nvpn_interfaces = get_vpn_interfaces()\nwireless_interfaces = {}\ninterfaces = {}\ndns_servers = {}\ndns_search = {}\n\nfunction new_address_array(address, interface, family)\n    proto = interface['proto']\n    if proto == 'dhcpv6' then\n        proto = 'dhcp'\n    end\n    new_address = {\n        address = address['address'],\n        mask = address['mask'],\n        proto = proto,\n        family = family,\n        gateway = find_default_gateway(interface.route)\n    }\n    return new_address\nend\n\nspecialized_interfaces = {\n    modemmanager = function(name, interface)\n        local modem = uci_cursor.get('network', interface['interface'], 'device')\n        local info = {}\n\n        local general = io.popen('mmcli --output-json -m '..modem):read(\"*a\")\n        if general and pcall(function () general = cjson.decode(general) end) then\n            general = general.modem\n\n            if not is_table_empty(general['3gpp']) then\n                info.imei = general['3gpp'].imei\n                info.operator_name = general['3gpp']['operator-name']\n                info.operator_code = general['3gpp']['operator-code']\n            end\n\n            if not is_table_empty(general.generic) then\n                info.manufacturer = general.generic.manufacturer\n                info.model = general.generic.model\n                info.connection_status = general.generic.state\n                info.power_status = general.generic['power-state']\n            end\n        end\n\n        local signal = io.popen('mmcli --output-json -m '..modem..' --signal-get'):read()\n        if signal and pcall(function () signal = cjson.decode(signal) end) then\n            -- only send data if not empty to avoid generating too much traffic\n            if not is_table_empty(signal.modem) and not is_table_empty(signal.modem.signal) then\n                -- omit refresh rate\n                signal.modem.signal.refresh = nil\n                info.signal = {}\n                -- collect section and values only if not empty\n                for section_key, section_values in pairs(signal.modem.signal) do\n                    for key, value in pairs(section_values) do\n                        if value ~= '--' then\n                            if is_table_empty(info.signal[section_key]) then\n                                info.signal[section_key] = {}\n                            end\n                            info.signal[section_key][key] = tonumber(value)\n                        end\n                    end\n                end\n            end\n        end\n\n        return {type='modem-manager', mobile=info}\n    end\n}\n\nfunction get_interface_info(name, netjson_interface)\n    info = {\n        dns_search = nil,\n        dns_servers = nil\n    }\n    for _, interface in pairs(interface_data['interface']) do\n        if interface['l3_device'] == name then\n            if next(interface['dns-search']) then\n                info.dns_search = interface['dns-search']\n            end\n            if next(interface['dns-server']) then\n                info.dns_servers = interface['dns-server']\n            end\n            if netjson_interface.type == 'bridge' then\n                info.stp = uci_cursor.get('network', interface['interface'], 'stp') == '1'\n            end\n            -- collect specialized info if available\n            local specialized_info = specialized_interfaces[interface.proto]\n            if specialized_info then\n                info.specialized = specialized_info(name, interface)\n            end\n        end\n    end\n    return info\nend\n\nfunction array_concat(source, destination)\n    table.foreach(source, function(key, value)\n        table.insert(destination, value)\n    end)\nend\n\nfunction dict_merge(source, destination)\n    table.foreach(source, function(key, value)\n        destination[key] = value\n    end)\nend\n\n-- collect interface addresses\nfunction get_addresses(name)\n    addresses = {}\n    interface_list = interface_data['interface']\n    addresses_list = {}\n    for _, interface in pairs(interface_list) do\n        if interface['l3_device'] == name then\n            proto = interface['proto']\n            if proto == 'dhcpv6' then\n                proto = 'dhcp'\n            end\n            for _, address in pairs(interface['ipv4-address']) do\n                table.insert(addresses_list, address['address'])\n                new_address = new_address_array(address, interface, 'ipv4')\n                table.insert(addresses, new_address)\n            end\n            for _, address in pairs(interface['ipv6-address']) do\n                table.insert(addresses_list, address['address'])\n                new_address = new_address_array(address, interface, 'ipv6')\n                table.insert(addresses, new_address)\n            end\n        end\n    end\n    for i = 1, #nixio_data do\n        if nixio_data[i].name == name then\n            if not is_excluded(name) then\n                family = nixio_data[i].family\n                addr = nixio_data[i].addr\n                if family == 'inet' then\n                    family = 'ipv4'\n                    -- Since we don't already know this from the dump, we can\n                    -- consider this dynamically assigned, this is the case for\n                    -- example for OpenVPN interfaces, which get their address\n                    -- from the DHCP server embedded in OpenVPN\n                    proto = 'dhcp'\n                elseif family == 'inet6' then\n                    family = 'ipv6'\n                    if starts_with(addr, 'fe80') then\n                        proto = 'static'\n                    else\n                        ula = uci_cursor.get('network', 'globals', 'ula_prefix')\n                        ula_prefix = split(ula, '::')[1]\n                        if starts_with(addr, ula_prefix) then\n                            proto = 'static'\n                        else\n                            proto = 'dhcp'\n                        end\n                    end\n                end\n                if family == 'ipv4' or family == 'ipv6' then\n                    if not has_value(addresses_list, addr) then\n                        table.insert(addresses, {\n                            address = addr,\n                            mask = nixio_data[i].prefix,\n                            proto = proto,\n                            family = family\n                        })\n                    end\n                end\n            end\n        end\n    end\n    return addresses\nend\n\n-- collect relevant wireless interface stats\n-- (traffic and connected clients)\nfor radio_name, radio in pairs(wireless_status) do\n    for i, interface in ipairs(radio.interfaces) do\n        name = interface.ifname\n        local is_mesh = false\n        if name and not is_excluded(name) then\n            iwinfo = ubus:call('iwinfo', 'info', {\n                device = name\n            })\n            netjson_interface = {\n                name = name,\n                type = 'wireless',\n                wireless = {\n                    ssid = iwinfo.ssid,\n                    mode = iwinfo_modes[iwinfo.mode] or iwinfo.mode,\n                    channel = iwinfo.channel,\n                    frequency = iwinfo.frequency,\n                    tx_power = iwinfo.txpower,\n                    signal = iwinfo.signal,\n                    noise = iwinfo.noise,\n                    country = iwinfo.country\n                }\n            }\n            if iwinfo.mode == 'Ad-Hoc' or iwinfo.mode == 'Mesh Point' then\n                clients = ubus:call('iwinfo', 'assoclist', {\n                    device = name\n                }).results\n                is_mesh = true\n            else\n              hostapd_output = ubus:call('hostapd.' .. name, 'get_clients', {})\n              if hostapd_output then\n                  clients = hostapd_output.clients\n              end\n            end\n            if clients and next(clients) ~= nil then\n                netjson_interface.wireless.clients = netjson_clients(clients, is_mesh)\n            end\n            wireless_interfaces[name] = netjson_interface\n        end\n    end\nend\n\nfunction needs_inversion(interface)\n    return interface.type == 'wireless' and interface.wireless.mode == 'access_point'\nend\n\nfunction invert_rx_tx(interface)\n    for k, v in pairs(interface) do\n        if string.sub(k, 0, 3) == \"rx_\" then\n            local tx_key = \"tx_\" .. string.sub(k, 4)\n            local tx_val = interface[tx_key]\n            interface[tx_key] = v\n            interface[k] = tx_val\n        end\n    end\n    return interface\nend\n\n-- collect interface stats\nfor name, interface in pairs(network_status) do\n    -- only collect data from iterfaces which have not been excluded\n    if not is_excluded(name) then\n        netjson_interface = {\n            name = name,\n            type = string.lower(interface.type),\n            up = interface.up,\n            mac = interface.macaddr,\n            txqueuelen = interface.txqueuelen,\n            mtu = interface.mtu,\n            speed = interface.speed,\n            bridge_members = interface['bridge-members'],\n            multicast = interface.multicast\n        }\n        if wireless_interfaces[name] then\n            dict_merge(wireless_interfaces[name], netjson_interface)\n            interface.type = netjson_interface.type\n        end\n        if interface.type == 'Network device' then\n            link_supported = interface['link-supported']\n            if link_supported and next(link_supported) then\n                netjson_interface.type = 'ethernet'\n                netjson_interface.link_supported = link_supported\n            elseif vpn_interfaces[name] then\n                netjson_interface.type = 'virtual'\n            else\n                netjson_interface.type = 'other'\n            end\n        end\n        if include_stats[name] or traffic_monitored == '*' then\n            if needs_inversion(netjson_interface) then\n                --- ensure wifi access point interfaces\n                --- show download and upload values from\n                --- the user's perspective and not from the router perspective\n                interface.statistics = invert_rx_tx(interface.statistics)\n            end\n            netjson_interface.statistics = interface.statistics\n        end\n        addresses = get_addresses(name)\n        if next(addresses) then\n            netjson_interface.addresses = addresses\n        end\n        info = get_interface_info(name, netjson_interface)\n        if info.stp ~= nil then\n            netjson_interface.stp = info.stp\n        end\n        if info.specialized then\n            for key, value in pairs(info.specialized) do\n                netjson_interface[key] = value\n            end\n        end\n        table.insert(interfaces, netjson_interface)\n        -- DNS info is independent from interface\n        if info.dns_servers then\n            array_concat(info.dns_servers, dns_servers)\n        end\n        if info.dns_search then\n            array_concat(info.dns_search, dns_search)\n        end\n    end\nend\n\nif next(interfaces) ~= nil then\n    netjson.interfaces = interfaces\nend\nif next(dns_servers) ~= nil then\n    netjson.dns_servers = dns_servers\nend\nif next(dns_search) ~= nil then\n    netjson.dns_search = dns_search\nend\n\nprint(cjson.encode(netjson))",  # noqa
    }
)
TEMPLATE_CRONTAB_MONITORING_01 = OrderedDict(
    {
        "path": "/etc/crontabs/root",
        "mode": "0644",
        "contents": "*/5 * * * * /usr/sbin/openwisp-monitoring\n",
    }
)
TEMPLATE_CRONTAB_MONITORING_02 = OrderedDict(
    {
        "path": "/etc/crontabs/root",
        "mode": "0644",
        "contents": "*/5 * * * * /usr/sbin/legacy-openwisp-monitoring\n",
    }
)
TEMPLATE_RC_LOCAL_01 = OrderedDict(
    {
        "path": "/etc/rc.local",
        "mode": "0644",
        "contents": "# Put your custom commands here that should be executed once\n# the system init finished. By default this file does nothing.\n\n/usr/sbin/openwisp-monitoring\ntouch /etc/crontabs/root\n/etc/init.d/cron start\n\nexit 0\n",  # noqa
    }
)
TEMPLATE_RC_LOCAL_02 = OrderedDict(
    {
        "path": "/etc/rc.local",
        "mode": "0644",
        "contents": "# Put your custom commands here that should be executed once\n# the system init finished. By default this file does nothing.\n\n/usr/sbin/legacy-openwisp-monitoring\ntouch /etc/crontabs/root\n/etc/init.d/cron start\n\nexit 0\n",  # noqa
    }
)
TEMPLATE_UPDATE_OPENWISP_PACKAGES_01 = OrderedDict(
    {
        "path": "/usr/sbin/update-openwisp-packages",
        "mode": "0744",
        "contents": "#!/bin/sh\nreboot=false\n\n# updates opkg lists only if necessary\nopkg_update(){\n    (\n        test -d /tmp/opkg-lists/ && \\\n        test -f /tmp/opkg-lists/openwrt_base && \\\n        test -f /tmp/opkg-lists/openwrt_packages && \\\n        test -f /tmp/opkg-lists/openwrt_core\n    ) || opkg update;\n}\n\n# installs libubus-lua if necessary\nlibubus_lua_installed=$(opkg list-installed | grep libubus-lua -c)\nif [ \"$libubus_lua_installed\" == \"0\" ]; then\n    opkg_update\n    opkg install libubus-lua\nfi\n\n# installs lua-cjson if necessary\nluacjson_installed=$(opkg list-installed | grep lua-cjson -c)\nif [ \"$luacjson_installed\" == \"0\" ]; then\n    opkg_update\n    opkg install lua-cjson\nfi\n\n# installs rpcd-mod-iwinfo if necessary\nrpcd_mod_iwinfo_installed=$(opkg list-installed | grep rpcd-mod-iwinfo -c)\nif [ \"$rpcd_mod_iwinfo_installed\" == \"0\" ]; then\n    opkg_update\n    opkg install rpcd-mod-iwinfo\n    reboot=true \nfi\n\n# upgrades openwisp-config if necessary\nopenwisp_config_version=$(openwisp_config --version)\nif [ \"$openwisp_config_version\" != \"openwisp-config 0.5.0\" ]; then\n    # backup config just in case...\n    cp /etc/config/openwisp /etc/config/openwisp-backup\n    opkg_update\n    opkg install http://downloads.openwisp.io/openwisp-config/2021-01-07-162007/openwisp-config-mbedtls_0.5.0-1_all.ipk\n    # restore backup\n    mv /etc/config/openwisp-backup /etc/config/openwisp\n    # remove default conf\n    rm /etc/config/openwisp-opkg\n    /etc/init.d/openwisp_config restart\nfi\n\n# reboots if rpcd-mod-iwinfo has been installed\nif [ \"$reboot\" == \"true\" ]; then\n    sleep 5\n    reboot && exit\nfi\n",  # noqa
    }
)
TEMPLATE_UPDATE_OPENWISP_PACKAGES_02 = OrderedDict(
    {
        "path": "/usr/sbin/legacy-update-openwisp-package",
        "mode": "0744",
        "contents": "#!/bin/sh\nreboot=false\n\n# updates opkg lists only if necessary\nopkg_update(){\n    (\n        test -d /tmp/opkg-lists/ && \\\n        test -f /tmp/opkg-lists/openwrt_base && \\\n        test -f /tmp/opkg-lists/openwrt_packages && \\\n        test -f /tmp/opkg-lists/openwrt_core\n    ) || opkg update;\n}\n\n# installs libubus-lua if necessary\nlibubus_lua_installed=$(opkg list-installed | grep libubus-lua -c)\nif [ \"$libubus_lua_installed\" == \"0\" ]; then\n    opkg_update\n    opkg install libubus-lua\nfi\n\n# installs lua-cjson if necessary\nluacjson_installed=$(opkg list-installed | grep lua-cjson -c)\nif [ \"$luacjson_installed\" == \"0\" ]; then\n    opkg_update\n    opkg install lua-cjson\nfi\n\n# installs rpcd-mod-iwinfo if necessary\nrpcd_mod_iwinfo_installed=$(opkg list-installed | grep rpcd-mod-iwinfo -c)\nif [ \"$rpcd_mod_iwinfo_installed\" == \"0\" ]; then\n    opkg_update\n    opkg install rpcd-mod-iwinfo\n    reboot=true \nfi\n\n# reboots if rpcd-mod-iwinfo has been installed\nif [ \"$reboot\" == \"true\" ]; then\n    sleep 5\n    reboot && exit\nfi\n",  # noqa
    }
)
TEMPLATE_POST_RELOAD_HOOK_01 = OrderedDict(
    {
        "path": "/etc/openwisp/post-reload-hook",
        "mode": "0744",
        "contents": "#!/bin/sh\ntouch /etc/crontabs/root\n/etc/init.d/cron start\n/usr/sbin/update-openwisp-packages\n/usr/sbin/openwisp-monitoring\n",  # noqa
    }
)
TEMPLATE_POST_RELOAD_HOOK_02 = OrderedDict(
    {
        "path": "/etc/openwisp/post-reload-hook",
        "mode": "0744",
        "contents": "#!/bin/sh\ntouch /etc/crontabs/root\n/etc/init.d/cron start\n/usr/sbin/legacy-update-openwisp-packages\n/usr/sbin/legacy-openwisp-monitoring\n",  # noqa
    }
)


def assign_permissions_to_groups(apps, schema_editor):
    create_default_permissions(apps, schema_editor)
    operators_read_only_admins_manage = [
        'devicedata',
        'devicemonitoring',
        'wificlient',
        'wifisession',
    ]
    manage_operations = ['add', 'change', 'delete']
    Group = get_swapped_model(apps, 'openwisp_users', 'Group')

    try:
        admin = Group.objects.get(name='Administrator')
        operator = Group.objects.get(name='Operator')
    # consider failures custom cases
    # that do not have to be dealt with
    except Group.DoesNotExist:
        return

    for model_name in operators_read_only_admins_manage:
        try:
            permission = Permission.objects.get(codename='view_{}'.format(model_name))
            operator.permissions.add(permission.pk)
        except Permission.DoesNotExist:
            pass
        for operation in manage_operations:
            admin.permissions.add(
                Permission.objects.get(
                    codename='{}_{}'.format(operation, model_name)
                ).pk
            )



from .backends import timeseries_db

chart_query = timeseries_db.queries.chart_query
default_chart_query = timeseries_db.queries.default_chart_query
device_data_query = timeseries_db.queries.device_data_query

__all__ = ['timeseries_db', 'chart_query', 'default_chart_query', 'device_data_query']



# Manually Created

from collections import OrderedDict

from django.db import migrations

from . import (
    TEMPLATE_CRONTAB_MONITORING_02,
    TEMPLATE_MONITORING_UUID,
    TEMPLATE_NETJSON_MONITORING_02,
    TEMPLATE_OPENWISP_MONITORING_02,
    TEMPLATE_POST_RELOAD_HOOK_02,
    TEMPLATE_RC_LOCAL_02,
    TEMPLATE_UPDATE_OPENWISP_PACKAGES_02,
)


def migrate_data(apps, schema_editor):
    Template = apps.get_model('config', 'Template')
    try:
        template = Template.objects.get(pk=TEMPLATE_MONITORING_UUID)
        template.config = OrderedDict(
            {
                'files': [
                    TEMPLATE_OPENWISP_MONITORING_02,
                    TEMPLATE_NETJSON_MONITORING_02,
                    TEMPLATE_CRONTAB_MONITORING_02,
                    TEMPLATE_RC_LOCAL_02,
                    TEMPLATE_UPDATE_OPENWISP_PACKAGES_02,
                    TEMPLATE_POST_RELOAD_HOOK_02,
                ],
                'openwisp': [
                    OrderedDict(
                        {
                            'config_name': 'monitoring',
                            'config_value': 'monitoring',
                            'included_interfaces': '*',
                        }
                    )
                ],
            }
        )
        template.full_clean()
        template.save()
    except Template.DoesNotExist:
        pass


class Migration(migrations.Migration):
    dependencies = [
        ('device_monitoring', '0002_create_template'),
    ]

    operations = [
        migrations.RunPython(migrate_data, reverse_code=migrations.RunPython.noop)
    ]



import time
from datetime import timedelta

from django.core.cache import cache
from django.utils.timezone import now
from swapper import load_model

from openwisp_users.tests.utils import TestOrganizationMixin

from ...db import timeseries_db
from ...db.backends import TIMESERIES_DB
from ..configuration import (
    register_chart,
    register_metric,
    unregister_chart,
    unregister_metric,
)

start_time = now()
ten_minutes_ago = start_time - timedelta(minutes=10)
Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')

default_message = (
    '{notification.actor.name} for device [{notification.target}]'
    '({notification.target_link}) {notification.verb}.'
)

test_notification = {
    'problem': {
        'verbose_name': 'Monitoring Alert',
        'verb': 'crossed the threshold',
        'level': 'warning',
        'email_subject': '[{site.name}] PROBLEM: {notification.actor.name} {notification.target}',
        'message': default_message,
    },
    'recovery': {
        'verbose_name': 'Monitoring Alert',
        'verb': 'returned within the threshold',
        'level': 'info',
        'email_subject': '[{site.name}] RECOVERY: {notification.actor.name} {notification.target}',
        'message': default_message,
    },
}

# these custom metric configurations are used for automated testing purposes
metrics = {
    'test_metric': {
        'name': 'dummy',
        'key': '{key}',
        'field_name': '{field_name}',
        'label': 'Test Metric',
        'notification': test_notification,
    },
    'test_alert_field': {
        'name': 'test_alert_related',
        'key': '{key}',
        'field_name': 'test_alert_field',
        'label': 'Test alert related',
        'notification': test_notification,
        'related_fields': ['test_related_1', 'test_related_2', 'test_related_3'],
        'alert_field': 'test_related_2',
    },
    'top_fields_mean': {
        'name': 'top_fields_mean_test',
        'key': '{key}',
        'field_name': '{field_name}',
        'label': 'top fields mean test',
        'related_fields': ['google', 'facebook', 'reddit'],
    },
    'get_top_fields': {
        'name': 'get_top_fields_test',
        'key': '{key}',
        'field_name': '{field_name}',
        'label': 'get top fields test',
        'related_fields': ['http2', 'ssh', 'udp', 'spdy'],
    },
}

# this custom chart configuration is used for automated testing purposes
charts = {
    'histogram': {
        'type': 'histogram',
        'title': 'Histogram',
        'description': 'Histogram',
        'top_fields': 2,
        'order': 999,
        'query': {
            'influxdb': (
                "SELECT {fields|SUM|/ 1} FROM {key} "
                "WHERE time >= '{time}' AND content_type = "
                "'{content_type}' AND object_id = '{object_id}'"
            )
        },
    },
    'dummy': {
        'type': 'line',
        'title': 'Dummy chart',
        'description': 'Dummy chart for testing purposes.',
        'unit': 'candies',
        'order': 999,
        'query': None,
    },
    'bad_test': {
        'type': 'line',
        'title': 'Bugged chart for testing purposes',
        'description': 'Bugged chart for testing purposes.',
        'unit': 'bugs',
        'order': 999,
        'query': {'influxdb': "BAD"},
    },
    'default': {
        'type': 'line',
        'title': 'Default query for testing purposes',
        'description': 'Default query for testing purposes',
        'unit': 'n.',
        'order': 999,
        'query': {
            'influxdb': (
                "SELECT {field_name} FROM {key} WHERE time >= '{time}' AND "
                "content_type = '{content_type}' AND object_id = '{object_id}'"
            )
        },
    },
    'multiple_test': {
        'type': 'line',
        'title': 'Multiple test',
        'description': 'For testing purposes',
        'unit': 'n.',
        'order': 999,
        'query': {
            'influxdb': (
                "SELECT {field_name}, value2 FROM {key} WHERE time >= '{time}' AND "
                "content_type = '{content_type}' AND object_id = '{object_id}'"
            )
        },
    },
    'group_by_tag': {
        'type': 'stackedbars',
        'title': 'Group by tag',
        'description': 'Query is groupped by tag along with time',
        'unit': 'n.',
        'order': 999,
        'query': {
            'influxdb': (
                "SELECT CUMULATIVE_SUM(SUM({field_name})) FROM {key} WHERE time >= '{time}'"
                " GROUP BY time(1d), metric_num"
            )
        },
        'summary_query': {
            'influxdb': (
                "SELECT SUM({field_name}) FROM {key} WHERE time >= '{time}'"
                " GROUP BY time(30d), metric_num"
            )
        },
    },
    'mean_test': {
        'type': 'line',
        'title': 'Mean test',
        'description': 'For testing purposes',
        'unit': 'n.',
        'order': 999,
        'query': {
            'influxdb': (
                "SELECT MEAN({field_name}) AS {field_name} FROM {key} WHERE time >= '{time}' AND "
                "content_type = '{content_type}' AND object_id = '{object_id}'"
            )
        },
    },
    'sum_test': {
        'type': 'line',
        'title': 'Sum test',
        'description': 'For testing purposes',
        'unit': 'n.',
        'order': 999,
        'query': {
            'influxdb': (
                "SELECT SUM({field_name}) AS {field_name} FROM {key} WHERE time >= '{time}' AND "
                "content_type = '{content_type}' AND object_id = '{object_id}'"
            )
        },
    },
    'top_fields_mean': {
        'type': 'histogram',
        'title': 'Top fields mean test',
        'description': 'For testing purposes',
        'top_fields': 2,
        'order': 999,
        'query': {
            'influxdb': (
                "SELECT {fields|MEAN} FROM {key} "
                "WHERE time >= '{time}' AND content_type = "
                "'{content_type}' AND object_id = '{object_id}'"
            )
        },
    },
}


class TestMonitoringMixin(TestOrganizationMixin):
    ORIGINAL_DB = TIMESERIES_DB['NAME']
    TEST_DB = f'{ORIGINAL_DB}_test'

    @classmethod
    def setUpClass(cls):
        # By default timeseries_db.db shall connect to the database
        # defined in settings when apps are loaded. We don't want that while testing
        timeseries_db.db_name = cls.TEST_DB
        del timeseries_db.db
        del timeseries_db.dbs
        timeseries_db.create_database()
        for key, value in metrics.items():
            register_metric(key, value)
        for key, value in charts.items():
            register_chart(key, value)
        cache.clear()
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        timeseries_db.drop_database()
        for metric_name in metrics.keys():
            unregister_metric(metric_name)
        for key in charts.keys():
            unregister_chart(key)
        cache.clear()
        super().tearDownClass()

    def tearDown(self):
        timeseries_db.delete_metric_data()
        super().tearDown()

    def _create_general_metric(self, **kwargs):
        opts = {
            'name': 'test_metric',
            'is_healthy': True,  # backward compatibility with old tests
            'is_healthy_tolerant': True,  # backward compatibility with old tests
            'configuration': 'test_metric',
        }
        opts.update(kwargs)
        m = Metric(**opts)
        m.full_clean()
        m.save()
        return m

    def _create_object_metric(self, **kwargs):
        opts = kwargs.copy()
        if 'content_object' not in opts:
            opts['content_object'] = self._create_user()
        if 'is_healthy' not in kwargs:
            kwargs['is_healthy'] = True  # backward compatibility with old tests
        return self._create_general_metric(**opts)

    def _create_alert_settings(self, **kwargs):
        alert_s = AlertSettings(**kwargs)
        alert_s.full_clean()
        alert_s.save()
        return alert_s

    def _create_chart(self, metric=None, test_data=True, configuration='dummy'):
        m = metric or self._create_object_metric()
        if test_data:
            m.write(3, time=now() - timedelta(days=2))
            m.write(6, time=now() - timedelta(days=1))
            m.write(9, time=now())
        c = Chart(metric=m, configuration=configuration)
        c.full_clean()
        c.save()
        return c

    @property
    def _is_timeseries_udp_writes(self):
        return TIMESERIES_DB.get('OPTIONS', {}).get('udp_writes', False)

    def _read_chart_or_metric(self, obj, *args, **kwargs):
        if self._is_timeseries_udp_writes:
            time.sleep(0.12)
        return obj.read(*args, **kwargs)

    def _read_metric(self, metric, *args, **kwargs):
        return self._read_chart_or_metric(metric, *args, **kwargs)

    def _read_chart(self, chart, *args, **kwargs):
        return self._read_chart_or_metric(chart, *args, **kwargs)

    def _write_metric(self, metric, *args, **kwargs):
        metric.write(*args, **kwargs)
        if self._is_timeseries_udp_writes:
            time.sleep(0.12)



import swapper
from django.db import migrations

from openwisp_monitoring.check.settings import AUTO_CONFIG_CHECK, AUTO_PING
from openwisp_monitoring.check.tasks import auto_create_config_check, auto_create_ping


def create_ping_checks(apps, schema_editor):
    if AUTO_PING:
        ContentType = apps.get_model('contenttypes', 'ContentType')
        Check = apps.get_model('check', 'Check')
        Device = apps.get_model('config', 'Device')
        for device in Device.objects.iterator():
            auto_create_ping(
                model=Device.__name__.lower(),
                app_label=Device._meta.app_label,
                object_id=str(device.pk),
                check_model=Check,
                content_type_model=ContentType,
            )


def create_config_applied_checks(apps, schema_editor):
    if not AUTO_CONFIG_CHECK:
        return
    ContentType = apps.get_model('contenttypes', 'ContentType')
    Check = apps.get_model('check', 'Check')
    Device = apps.get_model('config', 'Device')
    for device in Device.objects.iterator():
        auto_create_config_check(
            model=Device.__name__.lower(),
            app_label=Device._meta.app_label,
            object_id=str(device.pk),
            check_model=Check,
            content_type_model=ContentType,
        )


def remove_config_applied_checks(apps, schema_editor):
    Check = apps.get_model('check', 'Check')
    Metric = apps.get_model('monitoring', 'Metric')
    Check.objects.filter(
        check='openwisp_monitoring.check.classes.ConfigApplied'
    ).delete()
    Metric.objects.filter(configuration='config_applied').delete()


class Migration(migrations.Migration):
    dependencies = [
        ('check', '0006_rename_check_check_check_type'),
        swapper.dependency('monitoring', 'Metric'),
    ]

    operations = [
        migrations.RunPython(
            create_ping_checks, reverse_code=migrations.RunPython.noop
        ),
        migrations.RunPython(
            create_config_applied_checks, reverse_code=remove_config_applied_checks
        ),
    ]



import logging
from functools import reduce
from json import loads
from json.decoder import JSONDecodeError

from django.core.cache import cache
from django.core.exceptions import ValidationError
from jsonschema import draft7_format_checker, validate
from jsonschema.exceptions import ValidationError as SchemaError
from swapper import load_model

from openwisp_controller.connection.exceptions import NoWorkingDeviceConnectionError

from .. import settings as app_settings
from .base import BaseCheck

logger = logging.getLogger(__name__)

Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')
DeviceConnection = load_model('connection', 'DeviceConnection')

DEFAULT_IPERF3_CHECK_CONFIG = {
    'host': {
        'type': 'array',
        'items': {
            'type': 'string',
        },
        'default': [],
    },
    # username, password max_length chosen from iperf3 docs to avoid iperf3 param errors
    'username': {'type': 'string', 'default': '', 'minLength': 1, 'maxLength': 20},
    'password': {'type': 'string', 'default': '', 'minLength': 1, 'maxLength': 20},
    'rsa_public_key': {
        'type': 'string',
        'default': '',
    },
    'client_options': {
        'type': 'object',
        'properties': {
            'port': {
                'type': 'integer',
                'default': 5201,
                # max, min port chosen from iperf3 docs
                'minimum': 1,
                'maximum': 65535,
            },
            'time': {
                # sets the interval time in seconds
                # between periodic bandwidth, jitter, and loss reports
                'type': 'integer',
                'default': 10,
                'minimum': 1,
                # arbitrary chosen to avoid slowing down the queue (30min)
                'maximum': 1800,
            },
            'bytes': {
                # number of bytes to transmit (instead of 'time')
                'type': 'string',
                # default to '' since we're using time for
                # the test end condition instead of bytes
                'default': '',
            },
            'blockcount': {
                # number of blocks (packets) to transmit
                # instead of 'time' or 'bytes'
                'type': 'string',
                # default to '' since we're using time for
                # the test end condition instead of blockcount
                'default': '',
            },
            'window': {
                # window size / socket buffer size
                # this gets sent to the server and used on that side too
                'type': 'string',
                'default': '0',
            },
            'parallel': {
                # number of parallel client streams to run
                # note that iperf3 is single threaded
                # so if you are CPU bound this will not yield higher throughput
                'type': 'integer',
                'default': 1,
                # max, min parallel streams chosen from iperf3 docs
                'minimum': 1,
                'maximum': 128,
            },
            'reverse': {
                # reverse the direction of a test
                # the server sends data to the client
                'type': 'boolean',
                'default': False,
            },
            'bidirectional': {
                # test in both directions (normal and reverse)
                # with both the client and server sending
                # and receiving data simultaneously
                'type': 'boolean',
                'default': False,
            },
            'connect_timeout': {
                # set timeout for establishing the initial
                # control connection to the server, in milliseconds (ms)
                # providing a shorter value (ex. 1000 ms (1 sec)) may
                # speed up detection of a down iperf3 server
                'type': 'integer',
                'default': 1000,
                'minimum': 1,
                # arbitrary chosen to avoid slowing down the queue (1000 sec)
                'maximum': 1000000,
            },
            'tcp': {
                'type': 'object',
                'properties': {
                    'bitrate': {
                        # set target bitrate to n bits/sec
                        'type': 'string',
                        'default': '0',
                    },
                    'length': {
                        # length of buffer to read or write
                        'type': 'string',
                        # for TCP tests, the default value is 128KB
                        'default': '128K',
                    },
                },
            },
            'udp': {
                'type': 'object',
                'properties': {
                    'bitrate': {
                        'type': 'string',
                        # set target bitrate to n bits/sec
                        # 30 Mbps
                        'default': '30M',
                    },
                    'length': {
                        # iperf3 tries to dynamically determine a
                        # reasonable sending size based on the path MTU
                        # if that cannot be determined it uses 1460 bytes
                        'type': 'string',
                        'default': '0',
                    },
                },
            },
        },
    },
}


def get_iperf3_schema():
    schema = {
        '$schema': 'http://json-schema.org/draft-07/schema#',
        'type': 'object',
        'additionalProperties': True,
        'dependencies': {
            'username': ['password', 'rsa_public_key'],
            'password': ['username', 'rsa_public_key'],
            'rsa_public_key': ['username', 'password'],
        },
    }
    schema['properties'] = DEFAULT_IPERF3_CHECK_CONFIG
    return schema


class Iperf3(BaseCheck):
    schema = get_iperf3_schema()

    def validate_params(self, params=None):
        try:
            if not params:
                params = self.params
            validate(params, self.schema, format_checker=draft7_format_checker)
        except SchemaError as e:
            message = 'Invalid param'
            path = '/'.join(e.path)
            if path:
                message = '{0} in "{1}"'.format(message, path)
            message = '{0}: {1}'.format(message, e.message)
            raise ValidationError({'params': message}) from e

    def _validate_iperf3_config(self, org):
        # if iperf3 config is present and validate it's params
        if app_settings.IPERF3_CHECK_CONFIG:
            self.validate_params(
                params=app_settings.IPERF3_CHECK_CONFIG.get(str(org.id))
            )

    def check(self, store=True):
        lock_acquired = False
        org = self.related_object.organization
        self._validate_iperf3_config(org)
        available_iperf3_servers = self._get_param('host', 'host.default')
        if not available_iperf3_servers:
            logger.warning(
                (
                    f'Iperf3 servers for organization "{org}" '
                    f'is not configured properly, iperf3 check skipped!'
                )
            )
            return
        # Avoid running the iperf3 check when the device monitoring status is "critical"
        if (
            self.related_object.monitoring
            and self.related_object.monitoring.status == 'critical'
        ):
            logger.info(
                (
                    f'"{self.related_object}" DeviceMonitoring '
                    'health status is "critical", iperf3 check skipped!'
                )
            )
            return
        time = self._get_param(
            'client_options.time', 'client_options.properties.time.default'
        )
        # Try to acquire a lock, or put task back on queue
        for server in available_iperf3_servers:
            server_lock_key = f'ow_monitoring_{org}_iperf3_check_{server}'
            # Set available_iperf3_server to the org device
            lock_acquired = cache.add(
                server_lock_key,
                str(self.related_object),
                timeout=app_settings.IPERF3_CHECK_LOCK_EXPIRE,
            )
            if lock_acquired:
                break
        else:
            logger.info(
                (
                    f'At the moment, all available iperf3 servers of organization "{org}" '
                    f'are busy running checks, putting "{self.check_instance}" back in the queue..'
                )
            )
            # Return the iperf3_check task to the queue,
            # it will executed after 2 * iperf3_check_time (TCP+UDP)
            self.check_instance.perform_check_delayed(duration=2 * time)
            return
        try:
            # Execute the iperf3 check with current available server
            result = self._run_iperf3_check(store, server, time)
        finally:
            # Release the lock after completion of the check
            cache.delete(server_lock_key)
            return result

    def _run_iperf3_check(self, store, server, time):
        try:
            device_connection = DeviceConnection.get_working_connection(
                self.related_object
            )
        except NoWorkingDeviceConnectionError:
            logger.warning(
                f'Failed to get a working DeviceConnection for "{self.related_object}", iperf3 check skipped!'
            )
            return
        command_tcp, command_udp = self._get_check_commands(server)

        # TCP mode
        result, exit_code = device_connection.connector_instance.exec_command(
            command_tcp, raise_unexpected_exit=False
        )
        # Exit code 127 : command doesn't exist
        if exit_code == 127:
            logger.warning(
                f'Iperf3 is not installed on the "{self.related_object}", error - {result.strip()}'
            )
            return

        result_tcp = self._get_iperf3_result(result, exit_code, mode='TCP')
        # UDP mode
        result, exit_code = device_connection.connector_instance.exec_command(
            command_udp, raise_unexpected_exit=False
        )
        result_udp = self._get_iperf3_result(result, exit_code, mode='UDP')
        result = {}
        if store and result_tcp and result_udp:
            # Store iperf3_result field 1 if any mode passes, store 0 when both fails
            iperf3_result = result_tcp['iperf3_result'] | result_udp['iperf3_result']
            result.update({**result_tcp, **result_udp, 'iperf3_result': iperf3_result})
            self.store_result(result)
        device_connection.disconnect()
        return result

    def _get_check_commands(self, server):
        """Returns tcp & udp commands for iperf3 check."""
        username = self._get_param('username', 'username.default')
        port = self._get_param(
            'client_options.port', 'client_options.properties.port.default'
        )
        window = self._get_param(
            'client_options.window', 'client_options.properties.window.default'
        )
        parallel = self._get_param(
            'client_options.parallel', 'client_options.properties.parallel.default'
        )
        ct = self._get_param(
            'client_options.connect_timeout',
            'client_options.properties.connect_timeout.default',
        )
        tcp_bitrate = self._get_param(
            'client_options.tcp.bitrate',
            'client_options.properties.tcp.properties.bitrate.default',
        )
        tcp_length = self._get_param(
            'client_options.tcp.length',
            'client_options.properties.tcp.properties.length.default',
        )
        udp_bitrate = self._get_param(
            'client_options.udp.bitrate',
            'client_options.properties.udp.properties.bitrate.default',
        )
        udp_length = self._get_param(
            'client_options.udp.length',
            'client_options.properties.udp.properties.length.default',
        )

        rev_or_bidir, test_end_condition = self._get_iperf3_test_conditions()
        logger.info(f'«« Iperf3 server : {server}, Device : {self.related_object} »»')
        command_tcp = (
            f'iperf3 -c {server} -p {port} {test_end_condition} --connect-timeout {ct} '
            f'-b {tcp_bitrate} -l {tcp_length} -w {window} -P {parallel} {rev_or_bidir} -J'
        )
        command_udp = (
            f'iperf3 -c {server} -p {port} {test_end_condition} --connect-timeout {ct} '
            f'-b {udp_bitrate} -l {udp_length} -w {window} -P {parallel} {rev_or_bidir} -u -J'
        )

        # All three parameters ie. username, password and rsa_public_key is required
        # for authentication to work, checking only username here
        if username:
            password = self._get_param('password', 'password.default')
            key = self._get_param('rsa_public_key', 'rsa_public_key.default')
            rsa_public_key = self._get_compelete_rsa_key(key)
            rsa_public_key_path = '/tmp/iperf3-public-key.pem'

            command_tcp = (
                f'echo "{rsa_public_key}" > {rsa_public_key_path} && '
                f'IPERF3_PASSWORD="{password}" iperf3 -c {server} -p {port} {test_end_condition} '
                f'--username "{username}" --rsa-public-key-path {rsa_public_key_path} --connect-timeout {ct} '
                f'-b {tcp_bitrate} -l {tcp_length} -w {window} -P {parallel} {rev_or_bidir} -J'
            )

            command_udp = (
                f'IPERF3_PASSWORD="{password}" iperf3 -c {server} -p {port} {test_end_condition} '
                f'--username "{username}" --rsa-public-key-path {rsa_public_key_path} --connect-timeout {ct} '
                f'-b {udp_bitrate} -l {udp_length} -w {window} -P {parallel} {rev_or_bidir} -u -J'
            )

            # If IPERF3_CHECK_DELETE_RSA_KEY, remove rsa_public_key from the device
            if app_settings.IPERF3_CHECK_DELETE_RSA_KEY:
                command_udp = f'{command_udp} && rm -f {rsa_public_key_path}'
        return command_tcp, command_udp

    def _get_iperf3_test_conditions(self):
        """Returns iperf3 check test conditions (rev_or_bidir, end_condition)."""
        time = self._get_param(
            'client_options.time', 'client_options.properties.time.default'
        )
        bytes = self._get_param(
            'client_options.bytes', 'client_options.properties.bytes.default'
        )
        blockcount = self._get_param(
            'client_options.blockcount', 'client_options.properties.blockcount.default'
        )
        reverse = self._get_param(
            'client_options.reverse', 'client_options.properties.reverse.default'
        )
        bidirectional = self._get_param(
            'client_options.bidirectional',
            'client_options.properties.bidirectional.default',
        )
        # by default we use 'time' param
        # for the iperf3 test end condition
        test_end_condition = f'-t {time}'
        # if 'bytes' present in config
        # use it instead of 'time'
        if bytes:
            test_end_condition = f'-n {bytes}'
        # if 'blockcount' present in config
        # use it instead of 'time' or 'bytes'
        if blockcount:
            test_end_condition = f'-k {blockcount}'
        # only one reverse condition can be use
        # reverse or bidirectional not both
        rev_or_bidir = ''
        if reverse:
            rev_or_bidir = '--reverse'
        if bidirectional:
            rev_or_bidir = '--bidir'
        return rev_or_bidir, test_end_condition

    def _get_compelete_rsa_key(self, key):
        """Returns RSA key with proper format."""
        pem_prefix = '-----BEGIN PUBLIC KEY-----\n'
        pem_suffix = '\n-----END PUBLIC KEY-----'
        key = key.strip()
        return f'{pem_prefix}{key}{pem_suffix}'

    def _deep_get(self, dictionary, keys, default=None):
        """Returns dict key value using dict and it's dot_key string,

        ie: key1.key2_nested.key3_nested, if found, otherwise returns
        default.
        """
        return reduce(
            lambda d, key: d.get(key, default) if isinstance(d, dict) else default,
            keys.split("."),
            dictionary,
        )

    def _get_param(self, conf_key, default_conf_key):
        """Returns specified param or its default value according to the schema."""
        org_id = str(self.related_object.organization.id)
        iperf3_config = app_settings.IPERF3_CHECK_CONFIG

        if self.params:
            check_params = self._deep_get(self.params, conf_key)
            if check_params:
                return check_params

        if iperf3_config:
            iperf3_config = iperf3_config.get(org_id)
            iperf3_config_param = self._deep_get(iperf3_config, conf_key)
            if iperf3_config_param:
                return iperf3_config_param

        return self._deep_get(DEFAULT_IPERF3_CHECK_CONFIG, default_conf_key)

    def _get_iperf3_result(self, result, exit_code, mode):
        """Returns iperf3 test result."""
        try:
            result = loads(result)
        except JSONDecodeError:
            # Errors other than iperf3 test errors
            logger.warning(
                f'Iperf3 check failed for "{self.related_object}", error - {result.strip()}'
            )
            return

        if mode == 'TCP':
            if exit_code != 0:
                logger.warning(
                    f'Iperf3 check failed for "{self.related_object}", {result["error"]}'
                )
                return {
                    'iperf3_result': 0,
                    'sent_bps_tcp': 0.0,
                    'received_bps_tcp': 0.0,
                    'sent_bytes_tcp': 0,
                    'received_bytes_tcp': 0,
                    'retransmits': 0,
                }
            else:
                sent = result['end']['sum_sent']
                received = result['end']['sum_received']
                return {
                    'iperf3_result': 1,
                    'sent_bps_tcp': float(sent['bits_per_second']),
                    'received_bps_tcp': float(received['bits_per_second']),
                    'sent_bytes_tcp': sent['bytes'],
                    'received_bytes_tcp': received['bytes'],
                    'retransmits': sent['retransmits'],
                }

        elif mode == 'UDP':
            if exit_code != 0:
                logger.warning(
                    f'Iperf3 check failed for "{self.related_object}", {result["error"]}'
                )
                return {
                    'iperf3_result': 0,
                    'sent_bps_udp': 0.0,
                    'sent_bytes_udp': 0,
                    'jitter': 0.0,
                    'total_packets': 0,
                    'lost_packets': 0,
                    'lost_percent': 0.0,
                }
            else:
                return {
                    'iperf3_result': 1,
                    'sent_bps_udp': float(result['end']['sum']['bits_per_second']),
                    'sent_bytes_udp': result['end']['sum']['bytes'],
                    'jitter': float(result['end']['sum']['jitter_ms']),
                    'total_packets': result['end']['sum']['packets'],
                    'lost_packets': result['end']['sum']['lost_packets'],
                    'lost_percent': float(result['end']['sum']['lost_percent']),
                }

    def store_result(self, result):
        """Store result in the DB."""
        metric = self._get_metric()
        copied = result.copy()
        iperf3_result = copied.pop('iperf3_result')
        metric.write(iperf3_result, extra_values=copied)

    def _get_metric(self):
        """Gets or creates metric."""
        metric, created = self._get_or_create_metric()
        if created:
            self._create_alert_settings(metric)
            self._create_charts(metric)
        return metric

    def _create_alert_settings(self, metric):
        """Creates default iperf3 alert settings with is_active=False."""
        alert_settings = AlertSettings(metric=metric, is_active=False)
        alert_settings.full_clean()
        alert_settings.save()

    def _create_charts(self, metric):
        """Creates iperf3 related charts."""
        charts = [
            'bandwidth',
            'transfer',
            'retransmits',
            'jitter',
            'datagram',
            'datagram_loss',
        ]
        for chart in charts:
            chart = Chart(metric=metric, configuration=chart)
            chart.full_clean()
            chart.save()



from django.utils.translation import gettext_lazy as _
from django_filters import rest_framework as filters
from swapper import load_model

from openwisp_users.api.filters import (
    FilterDjangoByOrgManaged,
    OrganizationManagedFilter,
)

Device = load_model('config', 'Device')
WifiSession = load_model('device_monitoring', 'WifiSession')


class WifiSessionFilter(FilterDjangoByOrgManaged):
    class Meta:
        model = WifiSession
        fields = {
            'device__organization': ['exact'],
            'device': ['exact'],
            'device__group': ['exact'],
            'start_time': ['exact', 'gt', 'gte', 'lt', 'lte'],
            'stop_time': ['exact', 'gt', 'gte', 'lt', 'lte', 'isnull'],
        }


class MonitoringDeviceFilter(OrganizationManagedFilter):
    class Meta(OrganizationManagedFilter.Meta):
        model = Device
        fields = OrganizationManagedFilter.Meta.fields + ['monitoring__status']


class MonitoringNearbyDeviceFilter(OrganizationManagedFilter):
    distance__lte = filters.NumberFilter(
        label=_('Distance is less than or equal to'),
        field_name='distance',
        lookup_expr='lte',
    )
    model = filters.CharFilter(method='filter_model')

    class Meta(OrganizationManagedFilter.Meta):
        model = Device
        fields = OrganizationManagedFilter.Meta.fields + [
            'monitoring__status',
            'model',
        ]

    def filter_model(self, queryset, name, value):
        values = value.split('|')
        return queryset.filter(**{f"{name}__in": values})



from swapper import swappable_setting

from .base.models import AbstractAlertSettings, AbstractChart, AbstractMetric


class Metric(AbstractMetric):
    class Meta(AbstractMetric.Meta):
        abstract = False
        swappable = swappable_setting('monitoring', 'Metric')


class Chart(AbstractChart):
    class Meta(AbstractChart.Meta):
        abstract = False
        swappable = swappable_setting('monitoring', 'Chart')


class AlertSettings(AbstractAlertSettings):
    class Meta(AbstractAlertSettings.Meta):
        abstract = False
        swappable = swappable_setting('monitoring', 'AlertSettings')



# Generated by Django 3.2.18 on 2023-04-04 10:25

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('device_monitoring', '0005_add_group_permissions'),
    ]

    operations = [
        migrations.AlterField(
            model_name='wificlient',
            name='ht',
            field=models.BooleanField(
                blank=True, default=None, null=True, verbose_name='HT'
            ),
        ),
        migrations.AlterField(
            model_name='wificlient',
            name='vht',
            field=models.BooleanField(
                blank=True, default=None, null=True, verbose_name='VHT'
            ),
        ),
    ]



import json
import random
from collections import OrderedDict
from datetime import datetime

import swapper
from cache_memoize import cache_memoize
from dateutil.relativedelta import relativedelta
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from jsonschema import draft7_format_checker, validate
from jsonschema.exceptions import ValidationError as SchemaError
from model_utils import Choices
from model_utils.fields import StatusField
from netaddr import EUI, NotRegisteredError
from pytz import timezone as tz
from swapper import load_model

from openwisp_controller.config.validators import mac_address_validator
from openwisp_monitoring.device.settings import get_critical_device_metrics
from openwisp_utils.base import TimeStampedEditableModel

from ...db import device_data_query, timeseries_db
from ...monitoring.signals import threshold_crossed
from ...monitoring.tasks import _timeseries_write
from ...settings import CACHE_TIMEOUT
from .. import settings as app_settings
from .. import tasks
from ..schema import schema
from ..signals import health_status_changed
from ..utils import SHORT_RP, get_device_cache_key


def mac_lookup_cache_timeout():
    """Returns a random number of hours between 48 and 96.

    This avoids timing out the entire cache at the same time.
    """
    return 60 * 60 * random.randint(48, 96)


class AbstractDeviceData(object):
    schema = schema
    __data = None
    __key = 'device_data'
    __data_timestamp = None

    def __init__(self, *args, **kwargs):
        from ..writer import DeviceDataWriter

        self.data = kwargs.pop('data', None)
        self.writer = DeviceDataWriter(self)
        super().__init__(*args, **kwargs)

    @classmethod
    @cache_memoize(CACHE_TIMEOUT)
    def get_devicedata(cls, pk):
        obj = (
            cls.objects.select_related('devicelocation')
            .only(
                'id',
                'organization_id',
                'devicelocation__location_id',
                'devicelocation__floorplan_id',
            )
            .get(id=pk)
        )
        return obj

    @classmethod
    def invalidate_cache(cls, instance, *args, **kwargs):
        if isinstance(instance, load_model('geo', 'DeviceLocation')):
            pk = instance.content_object_id
        else:
            if kwargs.get('created'):
                return
            pk = instance.pk
        cls.get_devicedata.invalidate(cls, str(pk))

    def can_be_updated(self):
        """Do not attempt to push the conf if the device is not reachable."""
        can_be_updated = super().can_be_updated()
        return can_be_updated and self.monitoring.status not in ['critical', 'unknown']

    def _get_wifi_version(self, htmode):
        wifi_version_htmode = f'{_("Other")}: {htmode}'
        if 'NOHT' in htmode:
            wifi_version_htmode = f'{_("Legacy Mode")}: {htmode}'
        elif 'HE' in htmode:
            wifi_version_htmode = f'WiFi 6 (802.11ax): {htmode}'
        elif 'VHT' in htmode:
            wifi_version_htmode = f'WiFi 5 (802.11ac): {htmode}'
        elif 'HT' in htmode:
            wifi_version_htmode = f'WiFi 4 (802.11n): {htmode}'
        return wifi_version_htmode

    @property
    def data_user_friendly(self):
        if not self.data:
            return None
        data = self.data
        # slicing to eliminate the nanoseconds from timestamp
        measured_at = datetime.strptime(self.data_timestamp[0:19], '%Y-%m-%dT%H:%M:%S')
        time_elapsed = int((datetime.utcnow() - measured_at).total_seconds())
        if 'general' in data and 'local_time' in data['general']:
            local_time = data['general']['local_time']
            data['general']['local_time'] = datetime.fromtimestamp(
                local_time + time_elapsed, tz=tz('UTC')
            )
        if 'general' in data and 'uptime' in data['general']:
            uptime = '{0.days} days, {0.hours} hours and {0.minutes} minutes'
            data['general']['uptime'] = uptime.format(
                relativedelta(seconds=data['general']['uptime'] + time_elapsed)
            )
        # used for reordering interfaces
        interface_dict = OrderedDict()
        for interface in data.get('interfaces', []):
            # don't show interfaces if they don't have any useful info
            if len(interface.keys()) <= 2:
                continue
            # human readable wireless  mode
            if 'wireless' in interface and 'mode' in interface['wireless']:
                interface['wireless']['mode'] = interface['wireless']['mode'].replace(
                    '_', ' '
                )
            # convert to GHz
            if 'wireless' in interface and 'frequency' in interface['wireless']:
                interface['wireless']['frequency'] /= 1000
            # add wifi version
            if 'wireless' in interface and 'htmode' in interface['wireless']:
                interface['wireless']['htmode'] = self._get_wifi_version(
                    interface['wireless']['htmode']
                )

            interface_dict[interface['name']] = interface
        # reorder interfaces in alphabetical order
        interface_dict = OrderedDict(sorted(interface_dict.items()))
        data['interfaces'] = list(interface_dict.values())
        # reformat expiry in dhcp leases
        for lease in data.get('dhcp_leases', []):
            lease['expiry'] = datetime.fromtimestamp(lease['expiry'], tz=tz('UTC'))
        return data

    @property
    def data(self):
        """Retrieves last data snapshot from Timeseries Database."""
        if self.__data:
            return self.__data
        q = device_data_query.format(SHORT_RP, self.__key, self.pk)
        cache_key = get_device_cache_key(device=self, context='current-data')
        points = cache.get(cache_key)
        if not points:
            points = timeseries_db.get_list_query(q, precision=None)
        if not points:
            return None
        self.data_timestamp = points[0]['time']
        return json.loads(points[0]['data'])

    @data.setter
    def data(self, data):
        """Sets data."""
        self.__data = data

    @property
    def data_timestamp(self):
        """Retrieves timestamp at which the data was recorded."""
        return self.__data_timestamp

    @data_timestamp.setter
    def data_timestamp(self, value):
        """Sets the timestamp related to the data."""
        self.__data_timestamp = value

    def validate_data(self):
        """Validates data according to NetJSON DeviceMonitoring schema."""
        try:
            validate(self.data, self.schema, format_checker=draft7_format_checker)
        except SchemaError as e:
            path = [str(el) for el in e.path]
            trigger = '/'.join(path)
            message = 'Invalid data in "#/{0}", ' 'validator says:\n\n{1}'.format(
                trigger, e.message
            )
            raise ValidationError(message)

    def _transform_data(self):
        """Performs corrections or additions to the device data."""
        mac_detection = app_settings.MAC_VENDOR_DETECTION
        for interface in self.data.get('interfaces', []):
            # loop over mobile signal values to convert them to float
            if 'mobile' in interface and 'signal' in interface['mobile']:
                for signal_key, signal_values in interface['mobile']['signal'].items():
                    for key, value in signal_values.items():
                        signal_values[key] = float(value)
            # If HT/VHT/HE is not being used ie. htmode = 'NOHT',
            # set the HT/VHT/HE field of WiFi clients to None.
            # This is necessary because some clients may be
            # VHT capable but VHT is not enabled at the radio level,
            # which can mislead into thinking the client is not HT/VHT/HE capable.
            wireless = interface.get('wireless')
            if wireless and all(key in wireless for key in ('htmode', 'clients')):
                for client in wireless['clients']:
                    htmode = wireless['htmode']
                    ht_enabled = htmode.startswith('HT')
                    vht_enabled = htmode.startswith('VHT')
                    noht_enabled = htmode == 'NOHT'
                    if noht_enabled:
                        client['ht'] = client['vht'] = None
                        # since 'he' field is optional
                        if 'he' in client:
                            client['he'] = None
                    elif ht_enabled:
                        if client['vht'] is False:
                            client['vht'] = None
                        if client.get('he') is False:
                            client['he'] = None
                    elif vht_enabled and client.get('he') is False:
                        client['he'] = None
            # Convert bitrate from KBits/s to MBits/s
            if wireless and 'bitrate' in wireless:
                interface['wireless']['bitrate'] = round(
                    interface['wireless']['bitrate'] / 1000.0, 1
                )
            # add mac vendor to wireless clients if present
            if (
                not mac_detection
                or 'wireless' not in interface
                or 'clients' not in interface['wireless']
            ):
                continue
            for client in interface['wireless']['clients']:
                client['vendor'] = self._mac_lookup(client['mac'])
        if not mac_detection:
            return
        # add mac vendor to neighbors
        for neighbor in self.data.get('neighbors', []):
            # in some cases the mac_address may not be present
            # eg: neighbors with "FAILED" state
            neighbor['vendor'] = self._mac_lookup(neighbor.get('mac'))
        # add mac vendor to DHCP leases
        for lease in self.data.get('dhcp_leases', []):
            lease['vendor'] = self._mac_lookup(lease['mac'])

    @cache_memoize(mac_lookup_cache_timeout())
    def _mac_lookup(self, value):
        if not value:
            return ''
        try:
            return EUI(value).oui.registration().org
        except NotRegisteredError:
            return ''

    def save_data(self, time=None):
        """Validates and saves data to Timeseries Database."""
        self.validate_data()
        self._transform_data()
        time = time or now()
        options = dict(tags={'pk': self.pk}, timestamp=time, retention_policy=SHORT_RP)
        _timeseries_write(name=self.__key, values={'data': self.json()}, **options)
        cache_key = get_device_cache_key(device=self, context='current-data')
        # cache current data to allow getting it without querying the timeseries DB
        cache.set(
            cache_key,
            [
                {
                    'data': self.json(),
                    'time': time.astimezone(tz=tz('UTC')).isoformat(timespec='seconds'),
                }
            ],
            timeout=CACHE_TIMEOUT,
        )
        if app_settings.WIFI_SESSIONS_ENABLED:
            self.save_wifi_clients_and_sessions()

    def json(self, *args, **kwargs):
        return json.dumps(self.data, *args, **kwargs)

    def save_wifi_clients_and_sessions(self):
        _WIFICLIENT_FIELDS = ['vendor', 'ht', 'vht', 'he', 'wmm', 'wds', 'wps']
        WifiClient = load_model('device_monitoring', 'WifiClient')
        WifiSession = load_model('device_monitoring', 'WifiSession')

        active_sessions = []
        interfaces = self.data.get('interfaces', [])
        for interface in interfaces:
            if interface.get('type') != 'wireless':
                continue
            interface_name = interface.get('name')
            wireless = interface.get('wireless', {})
            if not wireless or wireless['mode'] != 'access_point':
                continue
            ssid = wireless.get('ssid')
            clients = wireless.get('clients', [])
            for client in clients:
                # Save WifiClient
                client_obj = WifiClient.get_wifi_client(client.get('mac'))
                update_fields = []
                for field in _WIFICLIENT_FIELDS:
                    if getattr(client_obj, field) != client.get(field):
                        setattr(client_obj, field, client.get(field))
                        update_fields.append(field)
                if update_fields:
                    client_obj.full_clean()
                    client_obj.save(update_fields=update_fields)

                # Save WifiSession
                session_obj, _ = WifiSession.objects.get_or_create(
                    device_id=self.id,
                    interface_name=interface_name,
                    ssid=ssid,
                    wifi_client=client_obj,
                    stop_time=None,
                )
                active_sessions.append(session_obj.pk)

        # Close open WifiSession
        WifiSession.objects.filter(
            device_id=self.id,
            stop_time=None,
        ).exclude(
            pk__in=active_sessions
        ).update(stop_time=now())


class AbstractDeviceMonitoring(TimeStampedEditableModel):
    device = models.OneToOneField(
        swapper.get_model_name('config', 'Device'),
        on_delete=models.CASCADE,
        related_name='monitoring',
    )
    STATUS = Choices(
        ('unknown', _(app_settings.HEALTH_STATUS_LABELS['unknown'])),
        ('ok', _(app_settings.HEALTH_STATUS_LABELS['ok'])),
        ('problem', _(app_settings.HEALTH_STATUS_LABELS['problem'])),
        ('critical', _(app_settings.HEALTH_STATUS_LABELS['critical'])),
        ('deactivated', _(app_settings.HEALTH_STATUS_LABELS['deactivated'])),
    )
    status = StatusField(
        _('health status'),
        db_index=True,
        help_text=_(
            '"{0}" means the device has been recently added; \n'
            '"{1}" means the device is operating normally; \n'
            '"{2}" means the device is having issues but it\'s still reachable; \n'
            '"{3}" means the device is not reachable or in critical conditions;\n'
            '"{4}" means the device is deactivated;'
        ).format(
            app_settings.HEALTH_STATUS_LABELS['unknown'],
            app_settings.HEALTH_STATUS_LABELS['ok'],
            app_settings.HEALTH_STATUS_LABELS['problem'],
            app_settings.HEALTH_STATUS_LABELS['critical'],
            app_settings.HEALTH_STATUS_LABELS['deactivated'],
        ),
    )

    class Meta:
        abstract = True

    def update_status(self, value):
        # don't trigger save nor emit signal if status is not changing
        if self.status == value:
            return
        self.status = value
        self.full_clean()
        self.save()
        # clear device management_ip when device is offline
        if self.status == 'critical' and app_settings.AUTO_CLEAR_MANAGEMENT_IP:
            self.device.management_ip = None
            self.device.save(update_fields=['management_ip'])

        health_status_changed.send(sender=self.__class__, instance=self, status=value)

    @property
    def related_metrics(self):
        Metric = load_model('monitoring', 'Metric')
        return Metric.objects.select_related('content_type').filter(
            object_id=self.device_id,
            content_type__model='device',
            content_type__app_label='config',
        )

    @staticmethod
    @receiver(threshold_crossed, dispatch_uid='threshold_crossed_receiver')
    def threshold_crossed(sender, metric, alert_settings, target, first_time, **kwargs):
        """Executed when a threshold is crossed.

        Changes the health status of a devicewhen a threshold defined in
        the alert settings related to the metric is crossed.
        """
        DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
        if not isinstance(target, DeviceMonitoring.device.field.related_model):
            return
        try:
            monitoring = target.monitoring
        except DeviceMonitoring.DoesNotExist:
            monitoring = DeviceMonitoring.objects.create(device=target)
        status = 'ok' if metric.is_healthy else 'problem'
        related_status = 'ok'
        for related_metric in monitoring.related_metrics.filter(is_healthy=False):
            if monitoring.is_metric_critical(related_metric):
                related_status = 'critical'
                break
            related_status = 'problem'
        if metric.is_healthy and related_status == 'problem':
            status = 'problem'
        elif metric.is_healthy and related_status == 'critical':
            status = 'critical'
        elif not metric.is_healthy and any(
            [monitoring.is_metric_critical(metric), related_status == 'critical']
        ):
            status = 'critical'
        monitoring.update_status(status)

    @staticmethod
    def is_metric_critical(metric):
        for critical in app_settings.CRITICAL_DEVICE_METRICS:
            if all(
                [
                    metric.key == critical['key'],
                    metric.field_name == critical['field_name'],
                ]
            ):
                return True
        return False

    @classmethod
    def handle_disabled_organization(cls, organization_id):
        """Handles the disabling of an organization.

        Clears the management IP of all devices belonging to a disabled
        organization and set their monitoring status to 'unknown'.

        Parameters: - organization_id (int): The ID of the disabled
        organization.

        Returns: - None
        """
        load_model('config', 'Device').objects.filter(
            organization_id=organization_id
        ).update(management_ip='')
        cls.objects.filter(device__organization_id=organization_id).update(
            status='unknown'
        )

    @classmethod
    def handle_deactivated_device(cls, instance, **kwargs):
        """Handles the deactivation of a device

        Sets the device's monitoring status to 'deactivated'

        Parameters: - instance (Device): The device object
        which is deactivated

        Returns: - None
        """
        cls.objects.filter(device_id=instance.id).update(status='deactivated')

    @classmethod
    def handle_activated_device(cls, instance, **kwargs):
        """Handles the activation of a deactivated device

        Sets the device's monitoring status to 'unknown'

        Parameters: - instance (Device): The device object
        which is deactivated

        Returns: - None
        """
        cls.objects.filter(device_id=instance.id).update(status='unknown')

    @classmethod
    def _get_critical_metric_keys(cls):
        return [metric['key'] for metric in get_critical_device_metrics()]

    @classmethod
    def handle_critical_metric(cls, instance, **kwargs):
        critical_metrics = cls._get_critical_metric_keys()
        if instance.check_type in critical_metrics:
            try:
                device_monitoring = cls.objects.get(device=instance.content_object)
                if not instance.is_active or kwargs.get('signal') == post_delete:
                    device_monitoring.update_status('unknown')
            except cls.DoesNotExist:
                pass


class AbstractWifiClient(TimeStampedEditableModel):
    id = None
    mac_address = models.CharField(
        max_length=17,
        db_index=True,
        primary_key=True,
        validators=[mac_address_validator],
        help_text=_('MAC address'),
    )
    vendor = models.CharField(max_length=200, blank=True, null=True)
    he = models.BooleanField(null=True, blank=True, default=None, verbose_name='HE')
    vht = models.BooleanField(null=True, blank=True, default=None, verbose_name='VHT')
    ht = models.BooleanField(null=True, blank=True, default=None, verbose_name='HT')
    wmm = models.BooleanField(default=False, verbose_name='WMM')
    wds = models.BooleanField(default=False, verbose_name='WDS')
    wps = models.BooleanField(default=False, verbose_name='WPS')

    class Meta:
        abstract = True
        verbose_name = _('WiFi Client')
        ordering = ('-created',)

    @classmethod
    @cache_memoize(CACHE_TIMEOUT)
    def get_wifi_client(cls, mac_address):
        wifi_client, _ = cls.objects.get_or_create(mac_address=mac_address)
        return wifi_client

    @classmethod
    def invalidate_cache(cls, instance, *args, **kwargs):
        if kwargs.get('created'):
            return
        cls.get_wifi_client.invalidate(cls, instance.mac_address)


class AbstractWifiSession(TimeStampedEditableModel):
    created = None

    device = models.ForeignKey(
        swapper.get_model_name('config', 'Device'),
        on_delete=models.CASCADE,
    )
    wifi_client = models.ForeignKey(
        swapper.get_model_name('device_monitoring', 'WifiClient'),
        on_delete=models.CASCADE,
    )
    ssid = models.CharField(
        max_length=32, blank=True, null=True, verbose_name=_('SSID')
    )
    interface_name = models.CharField(
        max_length=15,
    )
    start_time = models.DateTimeField(
        verbose_name=_('start time'),
        db_index=True,
        auto_now=True,
    )
    stop_time = models.DateTimeField(
        verbose_name=_('stop time'),
        db_index=True,
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True
        verbose_name = _('WiFi Session')
        ordering = ('-start_time',)

    def __str__(self):
        return self.mac_address

    @property
    def mac_address(self):
        return self.wifi_client.mac_address

    @property
    def vendor(self):
        return self.wifi_client.vendor

    @classmethod
    def offline_device_close_session(
        cls, metric, tolerance_crossed, first_time, target, **kwargs
    ):
        if (
            not first_time
            and tolerance_crossed
            and not metric.is_healthy_tolerant
            and AbstractDeviceMonitoring.is_metric_critical(metric)
        ):
            tasks.offline_device_close_session.delay(device_id=target.pk)



# Manually created
from django.db import migrations

from . import create_general_metrics, delete_general_metrics


class Migration(migrations.Migration):
    dependencies = [('monitoring', '0007_alter_metric_object_id')]

    operations = [
        migrations.RunPython(
            create_general_metrics, reverse_code=delete_general_metrics
        )
    ]



# Generated by Django 4.0.4 on 2022-05-17 11:28

import re
import uuid

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import swapper
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        swapper.dependency(
            *swapper.split(settings.AUTH_USER_MODEL), version='0004_default_groups'
        ),
        swapper.dependency('config', 'Device'),
        ('device_monitoring', '0003_update_template'),
    ]

    operations = [
        migrations.CreateModel(
            name='WifiClient',
            fields=[
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                (
                    'mac_address',
                    models.CharField(
                        db_index=True,
                        help_text='MAC address',
                        max_length=17,
                        primary_key=True,
                        serialize=False,
                        validators=[
                            django.core.validators.RegexValidator(
                                re.compile('^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$'),
                                code='invalid',
                                message='Must be a valid mac address.',
                            )
                        ],
                    ),
                ),
                ('vendor', models.CharField(blank=True, max_length=200, null=True)),
                ('ht', models.BooleanField(default=False, verbose_name='HT')),
                ('vht', models.BooleanField(default=False, verbose_name='VHT')),
                ('wmm', models.BooleanField(default=False, verbose_name='WMM')),
                ('wds', models.BooleanField(default=False, verbose_name='WDS')),
                ('wps', models.BooleanField(default=False, verbose_name='WPS')),
            ],
            options={
                'verbose_name': 'WiFi Client',
                'abstract': False,
                'swappable': 'DEVICE_MONITORING_WIFICLIENT_MODEL',
            },
        ),
        migrations.CreateModel(
            name='WifiSession',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                (
                    'ssid',
                    models.CharField(
                        blank=True, max_length=32, null=True, verbose_name='SSID'
                    ),
                ),
                ('interface_name', models.CharField(max_length=15)),
                (
                    'start_time',
                    models.DateTimeField(
                        auto_now=True, db_index=True, verbose_name='start time'
                    ),
                ),
                (
                    'stop_time',
                    models.DateTimeField(
                        blank=True, db_index=True, null=True, verbose_name='stop time'
                    ),
                ),
                (
                    'device',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('config', 'Device'),
                    ),
                ),
                (
                    'wifi_client',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('device_monitoring', 'WifiClient'),
                    ),
                ),
            ],
            options={
                'verbose_name': 'WiFi Session',
                'ordering': ('-start_time',),
                'abstract': False,
                'swappable': 'DEVICE_MONITORING_WIFISESSION_MODEL',
            },
        ),
    ]



from unittest.mock import patch

from django.urls import reverse
from openwisp_notifications.signals import notify
from swapper import load_model

from openwisp_controller.config.signals import config_status_changed
from openwisp_controller.connection.tests.base import CreateConnectionsMixin
from openwisp_utils.tests import catch_signal

from ...check.classes import Ping
from ...check.tests import _FPING_REACHABLE, _FPING_UNREACHABLE
from ..tasks import trigger_device_checks
from . import DeviceMonitoringTransactionTestcase

Check = load_model('check', 'Check')
DeviceMonitoring = load_model('device_monitoring', 'DeviceMonitoring')
DeviceConnection = load_model('connection', 'DeviceConnection')
Credentials = load_model('connection', 'Credentials')
Check = load_model('check', 'Check')


class TestTransactions(CreateConnectionsMixin, DeviceMonitoringTransactionTestcase):
    def _delete_non_ping_checks(self):
        Check.objects.exclude(name='Ping').delete()
        self.assertEqual(Check.objects.count(), 1)

    @patch('openwisp_monitoring.check.tasks.perform_check.delay')
    def test_config_status_changed_receiver(self, mock_method):
        c = self._create_config(status='applied', organization=self._create_org())
        c.config = {'general': {'description': 'test'}}
        c.full_clean()
        with catch_signal(config_status_changed) as handler:
            c.save()
            handler.assert_called_once()
        self.assertEqual(c.status, 'modified')
        self.assertEqual(mock_method.call_count, 1)

    @patch.object(Ping, '_command', return_value=_FPING_REACHABLE)
    def test_trigger_device_recovery_task(self, mocked_method):
        d = self._create_device(organization=self._create_org())
        d.management_ip = '10.40.0.5'
        d.save()
        data = self._data()
        # Creation of resources, clients and traffic metrics can be avoided here
        # as they are not involved. This speeds up the test by reducing requests made.
        del data['resources']
        del data['interfaces']
        self._delete_non_ping_checks()
        d.monitoring.update_status('critical')
        url = reverse('monitoring:api_device_metric', args=[d.pk.hex])
        url = '{0}?key={1}'.format(url, d.key)
        with patch.object(Check, 'perform_check') as mock:
            self._post_data(d.id, d.key, data)
            mock.assert_called_once()

    @patch.object(Ping, '_command', return_value=_FPING_UNREACHABLE)
    @patch.object(DeviceMonitoring, 'update_status')
    def test_trigger_device_recovery_task_regression(
        self, mocked_update_status, mocked_ping
    ):
        dm = self._create_device_monitoring()
        dm.device.management_ip = None
        dm.device.save()
        trigger_device_checks.delay(dm.device.pk)
        self.assertTrue(Check.objects.exists())
        # we expect update_status() to be called once (by the check)
        # and not a second time directly by our code
        mocked_update_status.assert_called_once()

    @patch.object(Check, 'perform_check')
    def test_is_working_false_true(self, perform_check):
        d = self._create_device()
        dm = d.monitoring
        dm.status = 'unknown'
        dm.save()
        self._delete_non_ping_checks()
        c = Credentials.objects.create()
        dc = DeviceConnection.objects.create(credentials=c, device=d, is_working=False)
        self.assertFalse(dc.is_working)
        dc.is_working = True
        dc.save()
        perform_check.assert_called_once()

    @patch.object(Check, 'perform_check')
    def test_is_working_changed_to_false(self, perform_check):
        d = self._create_device()
        dm = d.monitoring
        dm.status = 'ok'
        dm.save()
        self._delete_non_ping_checks()
        c = Credentials.objects.create()
        dc = DeviceConnection.objects.create(credentials=c, device=d)
        dc.is_working = False
        dc.save()
        perform_check.assert_called_once()

    @patch.object(Check, 'perform_check')
    @patch.object(notify, 'send')
    def test_is_working_none_true(self, notify_send, perform_check):
        d = self._create_device()
        dm = d.monitoring
        dm.status = 'unknown'
        dm.save()
        self._delete_non_ping_checks()
        c = Credentials.objects.create()
        dc = DeviceConnection.objects.create(credentials=c, device=d)
        self.assertIsNone(dc.is_working)
        dc.is_working = True
        dc.save()
        notify_send.assert_not_called()
        perform_check.assert_not_called()

    @patch.object(Check, 'perform_check')
    @patch.object(notify, 'send')
    def test_is_working_changed_unable_to_connect(self, notify_send, perform_check):
        ckey = self._create_credentials_with_key(port=self.ssh_server.port)
        dc = self._create_device_connection(credentials=ckey)
        dc.is_working = True
        dc.save()
        notify_send.assert_not_called()
        perform_check.assert_not_called()

        d = self.device_model.objects.first()
        d.monitoring.update_status('ok')
        self._delete_non_ping_checks()
        dc.is_working = False
        dc.failure_reason = '[Errno None] Unable to connect to port 5555 on 127.0.0.1'
        dc.full_clean()
        dc.save()
        perform_check.assert_not_called()

    @patch.object(Check, 'perform_check')
    @patch.object(notify, 'send')
    def test_is_working_changed_timed_out(self, notify_send, perform_check):
        ckey = self._create_credentials_with_key(port=self.ssh_server.port)
        dc = self._create_device_connection(credentials=ckey, is_working=None)
        self.assertIsNone(dc.is_working)
        dc.is_working = True
        dc.save()
        perform_check.assert_not_called()

        d = self.device_model.objects.first()
        d.monitoring.update_status('ok')
        self._delete_non_ping_checks()
        dc.is_working = False
        dc.failure_reason = 'timed out'
        dc.full_clean()
        dc.save()
        perform_check.assert_not_called()

    @patch.object(Check, 'perform_check')
    @patch.object(notify, 'send')
    def test_is_working_no_recovery_notification(self, notify_send, perform_check):
        ckey = self._create_credentials_with_key(port=self.ssh_server.port)
        dc = self._create_device_connection(credentials=ckey, is_working=True)
        d = self.device_model.objects.first()
        d.monitoring.update_status('ok')
        dc.refresh_from_db()
        self._delete_non_ping_checks()
        failure_reason = '[Errno None] Unable to connect to port 5555 on 127.0.0.1'
        self.assertTrue(dc.is_working)
        dc.failure_reason = failure_reason
        dc.is_working = False
        dc.save()
        # Recovery is made
        dc.failure_reason = ''
        dc.is_working = True
        dc.save()
        perform_check.assert_not_called()



from unittest.mock import patch

from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.test import tag
from django.urls.base import reverse
from reversion.models import Version
from selenium.common.exceptions import TimeoutException, UnexpectedAlertPresentException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from swapper import load_model

from openwisp_controller.connection.tests.utils import CreateConnectionsMixin
from openwisp_monitoring.device.tests import (
    TestDeviceMonitoringMixin,
    TestWifiClientSessionMixin,
)
from openwisp_monitoring.monitoring.configuration import DEFAULT_DASHBOARD_TRAFFIC_CHART
from openwisp_monitoring.monitoring.migrations import create_general_metrics
from openwisp_utils.admin_theme.dashboard import DASHBOARD_TEMPLATES
from openwisp_utils.test_selenium_mixins import (
    SeleniumTestMixin as BaseSeleniumTestMixin,
)

Device = load_model('config', 'Device')
DeviceConnection = load_model('connection', 'DeviceConnection')
DeviceData = load_model('device_monitoring', 'DeviceData')
AlertSettings = load_model('monitoring', 'AlertSettings')
Metric = load_model('monitoring', 'Metric')
Chart = load_model('monitoring', 'Chart')
Check = load_model('check', 'Check')


class SeleniumTestMixin(BaseSeleniumTestMixin):
    @classmethod
    def setUpClass(cls):
        """
        Sets up the necessary configurations for the test environment, ensuring
        that the dashboard templates render correctly during Selenium tests.

        During testing, the `OPENWISP_MONITORING_API_BASEURL` is set to
        `http://testserver`, a dummy value for the test environment. The dashboard
        templates are registered in the `AppConfig.ready` method, and these
        templates depend on specific URLs being correctly configured.

        Since mocking the `OPENWISP_MONITORING_API_BASEURL` does not update the
        URLs in the already registered dashboard templates, this method manually
        adjusts the template contexts to ensure they contain the correct URLs.
        """
        super().setUpClass()
        cls._dashboard_map_context = DASHBOARD_TEMPLATES[0][1].copy()
        cls._dashboard_timeseries_context = DASHBOARD_TEMPLATES[55][1].copy()
        DASHBOARD_TEMPLATES[0][1] = {
            'monitoring_device_list_url': reverse(
                'monitoring:api_location_device_list',
                args=['000'],
            ),
            'monitoring_location_geojson_url': reverse(
                'monitoring:api_location_geojson'
            ),
        }
        DASHBOARD_TEMPLATES[55][1]['api_url'] = reverse(
            'monitoring_general:api_dashboard_timeseries'
        )

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        DASHBOARD_TEMPLATES[0][1] = cls._dashboard_map_context
        DASHBOARD_TEMPLATES[55][1] = cls._dashboard_timeseries_context

    def setUp(self):
        self.admin = self._create_admin(
            username=self.admin_username, password=self.admin_password
        )
        super().setUp()


@tag("selenium_tests")
class TestDeviceConnectionInlineAdmin(
    SeleniumTestMixin,
    TestDeviceMonitoringMixin,
    CreateConnectionsMixin,
    StaticLiveServerTestCase,
):
    config_app_label = 'config'

    def tearDown(self):
        # Accept unsaved changes alert to allow other tests to run
        try:
            self.web_driver.refresh()
        except UnexpectedAlertPresentException:
            self.web_driver.switch_to_alert().accept()
        else:
            try:
                WebDriverWait(self.web_driver, 1).until(EC.alert_is_present())
            except TimeoutException:
                pass
            else:
                self.web_driver.switch_to_alert().accept()
        self.web_driver.refresh()
        WebDriverWait(self.web_driver, 2).until(
            EC.visibility_of_element_located((By.XPATH, '//*[@id="site-name"]'))
        )
        super().tearDown()

    def test_restoring_deleted_device(self):
        org = self._get_org()
        self._create_credentials(auto_add=True, organization=org)
        device = self._create_config(organization=org).device
        device_data = DeviceData.objects.get(id=device.id)
        device_checks = device_data.checks.all()
        for check in device_checks:
            check.perform_check()
        device_metric_ids = device_data.metrics.values_list('id', flat=True)
        device_chart_ids = Chart.objects.filter(
            metric_id__in=device_metric_ids
        ).values_list('id', flat=True)
        device_alert_setting_ids = AlertSettings.objects.filter(
            metric_id__in=device_metric_ids
        ).values_list('id', flat=True)
        self.assertEqual(len(device_alert_setting_ids), 2)
        self.assertEqual(len(device_metric_ids), 2)
        self.assertEqual(len(device_checks), 3)
        self.assertEqual(len(device_chart_ids), 3)

        self.login()

        # Save device to create revision.
        self.open(
            reverse(f'admin:{self.config_app_label}_device_change', args=[device.id])
        )
        self.web_driver.find_element(By.CSS_SELECTOR, 'input[name="_save"]').click()
        try:
            WebDriverWait(self.web_driver, 5).until(
                EC.url_to_be(f'{self.live_server_url}/admin/config/device/')
            )
        except TimeoutException:
            self.fail('Failed saving device')

        # Delete the device
        device.deactivate()
        device.config.set_status_deactivated()
        self.open(
            reverse(f'admin:{self.config_app_label}_device_delete', args=[device.id])
        )
        self.web_driver.find_element(
            by=By.CSS_SELECTOR, value='#content form input[type="submit"]'
        ).click()
        WebDriverWait(self.web_driver, 5).until(
            EC.url_to_be(
                '{}{}'.format(
                    self.live_server_url,
                    reverse(f'admin:{self.config_app_label}_device_changelist'),
                )
            )
        )
        self.assertEqual(Device.objects.count(), 0)
        self.assertEqual(DeviceConnection.objects.count(), 0)
        self.assertEqual(Check.objects.count(), 0)
        self.assertEqual(Metric.objects.exclude(object_id=None).count(), 0)

        version_obj = Version.objects.get_deleted(model=Device).first()

        # Restore deleted device
        self.open(
            reverse(
                f'admin:{self.config_app_label}_device_recover', args=[version_obj.id]
            )
        )
        self.web_driver.find_element(
            By.XPATH, '//*[@id="device_form"]/div/div[1]/input[1]'
        ).click()
        try:
            WebDriverWait(self.web_driver, 5).until(
                EC.url_to_be(f'{self.live_server_url}/admin/config/device/')
            )
        except TimeoutException:
            self.fail('Deleted device was not restored')

        self.assertEqual(Device.objects.count(), 1)
        self.assertEqual(DeviceConnection.objects.count(), 1)
        # Ensure that existing Metric, AlertSetting and Chart objects
        # was restored
        self.assertEqual(Metric.objects.filter(id__in=device_metric_ids).count(), 2)
        self.assertEqual(
            AlertSettings.objects.filter(id__in=device_alert_setting_ids).count(), 2
        )
        self.assertEqual(Chart.objects.filter(id__in=device_chart_ids).count(), 3)


@tag("selenium_tests")
class TestDashboardCharts(
    SeleniumTestMixin, TestDeviceMonitoringMixin, StaticLiveServerTestCase
):
    def setUp(self):
        super().setUp()
        # TransactionTestCase flushes the data, hence the metrics created
        # with migrations are lost. Only create general metrics if required.
        if Metric.objects.filter(object_id=None).count() != 2:
            create_general_metrics(None, None)
        self.assertEqual(Metric.objects.filter(object_id=None).count(), 2)

    @patch.dict(DEFAULT_DASHBOARD_TRAFFIC_CHART, {'__all__': ['wlan0', 'wlan1']})
    def test_dashboard_timeseries_charts(self):
        self.login()
        try:
            WebDriverWait(self.web_driver, 5).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, '#ow-chart-inner-container')
                )
            )
        except TimeoutException:
            self.fail('Timeseries chart container not found on dashboard')
        try:
            WebDriverWait(self.web_driver, 5).until(
                EC.visibility_of_element_located((By.CSS_SELECTOR, '#ow-chart-utils'))
            )
        except TimeoutException:
            self.fail('Timeseries chart time filter not found on dashboard')

        try:
            WebDriverWait(self.web_driver, 5).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, '#ow-chart-fallback')
                )
            )
        except TimeoutException:
            self.fail('Fallback message for charts did not render')
        else:
            self.assertIn(
                'Insufficient data for selected time period.',
                self.web_driver.find_element(
                    By.CSS_SELECTOR, '#ow-chart-fallback'
                ).get_attribute('innerHTML'),
            )
        self.create_test_data()
        self.web_driver.refresh()
        try:
            WebDriverWait(self.web_driver, 20).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, '#ow-chart-contents')
                )
            )
            WebDriverWait(self.web_driver, 20).until(
                EC.visibility_of_element_located((By.CSS_SELECTOR, '#chart-0'))
            )
            WebDriverWait(self.web_driver, 60).until(
                EC.visibility_of_element_located((By.CSS_SELECTOR, '#chart-1'))
            )
        except TimeoutException:
            self.fail('Timeseries charts did not render')

        self.assertIn(
            'General WiFi Clients',
            self.web_driver.find_element(
                By.CSS_SELECTOR, '#chart-0 > h3'
            ).get_attribute('innerHTML'),
        )
        self.assertIn(
            'General Traffic',
            self.web_driver.find_element(
                By.CSS_SELECTOR, '#chart-1 > h3'
            ).get_attribute('innerHTML'),
        )
        self.assertIn(
            'Open WiFi session list',
            self.web_driver.find_element(
                By.CSS_SELECTOR, '#chart-0-quick-link-container'
            ).get_attribute('innerHTML'),
        )


@tag("selenium_tests")
class TestWifiSessionInlineAdmin(
    SeleniumTestMixin,
    TestWifiClientSessionMixin,
    StaticLiveServerTestCase,
):
    config_app_label = 'config'

    def test_device_wifi_session_inline_change(self):
        dm = self._create_device_monitoring()
        device = dm.device
        self.login()
        path = f'admin:{self.config_app_label}_device_change'
        self.open(reverse(path, args=[device.pk]))
        # Make sure the wifi session inline doesn't exist
        WebDriverWait(self.web_driver, 2).until(
            EC.invisibility_of_element_located(
                (By.CSS_SELECTOR, '#wifisession_set-group')
            )
        )
        # We are still on the device change page,
        # and now we will create new wifi sessions
        ws1 = self._create_wifi_session(device=device)
        wc2 = self._create_wifi_client(mac_address='22:33:44:55:66:88')
        ws2 = self._create_wifi_session(
            device=device, wifi_client=wc2, ssid='Test Wifi Session'
        )
        WebDriverWait(self.web_driver, 10).until(
            EC.visibility_of_element_located(
                (By.CSS_SELECTOR, 'input[name="_continue"]')
            )
        )
        # Now press the 'Save' button on the device change page
        self.web_driver.find_element(By.CSS_SELECTOR, 'input[name="_continue"]').click()
        # Make sure the wifi session tab now
        # exists with the correct wifi sessions
        WebDriverWait(self.web_driver, 5).until(
            EC.visibility_of_element_located(
                (By.XPATH, '//*[@id="tabs-container"]/ul/li[7]/a')
            )
        )
        wifi_session_inline = self.web_driver.find_element(
            By.XPATH, '//*[@id="tabs-container"]/ul/li[7]/a'
        )
        wifi_session_inline.click()
        wifi_session_inline_form_error = (
            'ManagementForm data is missing '
            'or has been tampered with. Missing fields: '
            'wifisession_set-TOTAL_FORMS, wifisession_set-INITIAL_FORMS.'
        )
        # Make sure no inline formset errors
        # were encountered after saving
        self.assertNotIn(
            wifi_session_inline_form_error,
            self.web_driver.find_element(
                By.CSS_SELECTOR, '#wifisession_set-group'
            ).get_attribute('innerHTML'),
        )
        # Make sure all wifi sessions are present
        self.assertIn(
            f'{ws1.ssid}',
            self.web_driver.find_element(
                By.CSS_SELECTOR, '#wifisession_set-group'
            ).get_attribute('innerHTML'),
        )
        self.assertIn(
            f'{ws2.ssid}',
            self.web_driver.find_element(
                By.CSS_SELECTOR, '#wifisession_set-group'
            ).get_attribute('innerHTML'),
        )
        self.assertIn(
            'View Full History of WiFi Sessions',
            self.web_driver.find_element(
                By.CSS_SELECTOR, '#wifisession_set-group'
            ).get_attribute('innerHTML'),
        )



import logging
from copy import deepcopy
from datetime import datetime, timedelta

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from pytz import UTC
from swapper import load_model

from .. import settings as monitoring_settings
from ..monitoring.configuration import ACCESS_TECHNOLOGIES

Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
AlertSettings = load_model('monitoring', 'AlertSettings')
Device = load_model('config', 'Device')

logger = logging.getLogger(__name__)


class DeviceDataWriter(object):
    """This class is in charge of writing the device metric data.

    Before these methods were shipped in the REST API view but later have
    been moved here to allow writing this data in the background processes
    of OpenWISP.
    """

    def __init__(self, device_data):
        self.device_data = device_data

    def _init_previous_data(self):
        """makes NetJSON interfaces of previous snapshots more easy to access"""
        data = self.device_data.data or {}
        if data:
            data = deepcopy(data)
            data['interfaces_dict'] = {}
        for interface in data.get('interfaces', []):
            data['interfaces_dict'][interface['name']] = interface
        self._previous_data = data

    def _append_metric_data(
        self, metric, value, current=False, time=None, extra_values=None
    ):
        """Appends data for writing.

        Appends to the data structure which holds metric data and which
        will be sent to the timeseries DB.
        """
        self.write_device_metrics.append(
            (
                metric,
                {
                    'value': value,
                    'current': current,
                    'time': time,
                    'extra_values': extra_values,
                },
            )
        )

    def write(self, data, time=None, current=False):
        time = datetime.strptime(time, '%d-%m-%Y_%H:%M:%S.%f').replace(tzinfo=UTC)
        self._init_previous_data()
        self.device_data.data = data
        # saves raw device data
        self.device_data.save_data()
        data = self.device_data.data
        ct = ContentType.objects.get_for_model(Device)
        device_extra_tags = self._get_extra_tags(self.device_data)
        self.write_device_metrics = []
        for interface in data.get('interfaces', []):
            ifname = interface['name']
            if 'mobile' in interface:
                self._write_mobile_signal(
                    interface, ifname, ct, self.device_data.pk, current, time=time
                )
            ifstats = interface.get('statistics', {})
            # Explicitly stated None to avoid skipping in case the stats are zero
            if (
                ifstats.get('rx_bytes') is not None
                or ifstats.get('tx_bytes') is not None
            ):
                field_value = self._calculate_increment(
                    ifname, 'rx_bytes', ifstats.get('rx_bytes', 0)
                )
                extra_values = {
                    'tx_bytes': self._calculate_increment(
                        ifname, 'tx_bytes', ifstats.get('tx_bytes', 0)
                    )
                }
                name = f'{ifname} traffic'
                metric, created = Metric._get_or_create(
                    object_id=self.device_data.pk,
                    content_type_id=ct.id,
                    configuration='traffic',
                    name=name,
                    key='traffic',
                    main_tags={'ifname': Metric._makekey(ifname)},
                    extra_tags=device_extra_tags,
                )
                self._append_metric_data(
                    metric, field_value, current, time=time, extra_values=extra_values
                )
                if created:
                    self._create_traffic_chart(metric)
            try:
                clients = interface['wireless']['clients']
            except KeyError:
                continue
            if not isinstance(clients, list):
                continue
            name = '{0} wifi clients'.format(ifname)
            metric, created = Metric._get_or_create(
                object_id=self.device_data.pk,
                content_type_id=ct.id,
                configuration='clients',
                name=name,
                key='wifi_clients',
                main_tags={'ifname': Metric._makekey(ifname)},
                extra_tags=device_extra_tags,
            )
            # avoid tsdb overwrite clients
            client_time = time
            for client in clients:
                if 'mac' not in client:
                    continue
                self._append_metric_data(
                    metric, client['mac'], current, time=client_time
                )
                client_time += timedelta(microseconds=1)
            if created:
                self._create_clients_chart(metric)
        if 'resources' in data:
            if 'load' in data['resources'] and 'cpus' in data['resources']:
                self._write_cpu(
                    data['resources']['load'],
                    data['resources']['cpus'],
                    self.device_data.pk,
                    ct,
                    current,
                    time=time,
                )
            if 'disk' in data['resources']:
                self._write_disk(
                    data['resources']['disk'], self.device_data.pk, ct, time=time
                )
            if 'memory' in data['resources']:
                self._write_memory(
                    data['resources']['memory'],
                    self.device_data.pk,
                    ct,
                    current,
                    time=time,
                )
        try:
            Metric.batch_write(self.write_device_metrics)
        except ValueError as error:
            logger.error(
                f'Failed to write metrics for "{self.device_data.pk}" device.'
                f' Error: {error}'
            )

    def _get_extra_tags(self, device):
        tags = {'organization_id': str(device.organization_id)}
        try:
            device_location = device.devicelocation
        except ObjectDoesNotExist:
            pass
        else:
            tags['location_id'] = str(device_location.location_id)
            if device_location.floorplan_id:
                tags['floorplan_id'] = str(device_location.floorplan_id)
        return Metric._sort_dict(tags)

    def _get_mobile_signal_type(self, signal):
        if not signal:
            return
        access_techs = list(ACCESS_TECHNOLOGIES.keys())
        access_techs.reverse()
        # if only one access technology is in use, return that
        sections = list(signal.keys())
        if len(sections) == 1:
            return sections[0] if sections[0] in access_techs else None
        # if multiple mobile access technologies are in use,
        # return the most evolved technology in use
        for tech in access_techs:
            if tech in signal:
                return tech

    def _write_mobile_signal(self, interface, ifname, ct, pk, current=False, time=None):
        access_type = self._get_mobile_signal_type(interface['mobile'].get('signal'))
        if not access_type:
            return
        data = interface['mobile']['signal'][access_type]
        signal_power = signal_strength = None
        extra_values = {}
        if 'rssi' in data:
            signal_strength = data['rssi']
        if 'rsrp' in data:
            signal_power = data['rsrp']
        elif 'rscp' in data:
            signal_power = data['rscp']
        if signal_power is not None:
            extra_values = {'signal_power': float(signal_power)}
        if signal_strength is not None:
            signal_strength = float(signal_strength)
        if signal_strength is not None or signal_power is not None:
            metric, created = Metric._get_or_create(
                object_id=self.device_data.pk,
                content_type_id=ct.id,
                configuration='signal_strength',
                name='signal strength',
                key='signal',
                main_tags={'ifname': Metric._makekey(ifname)},
            )
            self._append_metric_data(
                metric, signal_strength, current, time=time, extra_values=extra_values
            )
            if created:
                self._create_signal_strength_chart(metric)

        snr = signal_quality = None
        extra_values = {}
        if 'snr' in data:
            snr = data['snr']
        if 'rsrq' in data:
            signal_quality = data['rsrq']
        if 'ecio' in data:
            snr = data['ecio']
        if 'sinr' in data:
            snr = data['sinr']
        if snr is not None:
            extra_values = {'snr': float(snr)}
        if signal_quality is not None:
            signal_quality = float(signal_quality)
        if snr is not None or signal_quality is not None:
            metric, created = Metric._get_or_create(
                object_id=self.device_data.pk,
                content_type_id=ct.id,
                configuration='signal_quality',
                name='signal quality',
                key='signal',
                main_tags={'ifname': Metric._makekey(ifname)},
            )
            self._append_metric_data(
                metric, signal_quality, current, time=time, extra_values=extra_values
            )
            if created:
                self._create_signal_quality_chart(metric)
        # create access technology chart
        metric, created = Metric._get_or_create(
            object_id=self.device_data.pk,
            content_type_id=ct.id,
            configuration='access_tech',
            name='access technology',
            key='signal',
            main_tags={'ifname': Metric._makekey(ifname)},
        )
        self._append_metric_data(
            metric,
            list(ACCESS_TECHNOLOGIES.keys()).index(access_type),
            current,
            time=time,
        )
        if created:
            self._create_access_tech_chart(metric)

    def _write_cpu(
        self, load, cpus, primary_key, content_type, current=False, time=None
    ):
        extra_values = {
            'load_1': float(load[0]),
            'load_5': float(load[1]),
            'load_15': float(load[2]),
        }
        metric, created = Metric._get_or_create(
            object_id=primary_key, content_type_id=content_type.id, configuration='cpu'
        )
        if created:
            self._create_resources_chart(metric, resource='cpu')
            self._create_resources_alert_settings(metric, resource='cpu')
        self._append_metric_data(
            metric,
            100 * float(load[0] / cpus),
            current,
            time=time,
            extra_values=extra_values,
        )

    def _write_disk(
        self, disk_list, primary_key, content_type, current=False, time=None
    ):
        used_bytes, size_bytes, available_bytes = 0, 0, 0
        for disk in disk_list:
            used_bytes += disk['used_bytes']
            size_bytes += disk['size_bytes']
            available_bytes += disk['available_bytes']
        metric, created = Metric._get_or_create(
            object_id=primary_key, content_type_id=content_type.id, configuration='disk'
        )
        if created:
            self._create_resources_chart(metric, resource='disk')
            self._create_resources_alert_settings(metric, resource='disk')
        self._append_metric_data(
            metric, 100 * used_bytes / size_bytes, current, time=time
        )

    def _write_memory(
        self, memory, primary_key, content_type, current=False, time=None
    ):
        extra_values = {
            'total_memory': memory['total'],
            'free_memory': memory['free'],
            'buffered_memory': memory['buffered'],
            'shared_memory': memory['shared'],
        }
        if 'cached' in memory:
            extra_values['cached_memory'] = memory.get('cached')
        percent_used = 100 * (
            1 - (memory['free'] + memory['buffered']) / memory['total']
        )
        # Available Memory is not shown in some systems (older openwrt versions)
        if 'available' in memory:
            extra_values.update({'available_memory': memory['available']})
            if memory['available'] > memory['free']:
                percent_used = 100 * (
                    1 - (memory['available'] + memory['buffered']) / memory['total']
                )
        metric, created = Metric._get_or_create(
            object_id=primary_key,
            content_type_id=content_type.id,
            configuration='memory',
        )
        if created:
            self._create_resources_chart(metric, resource='memory')
            self._create_resources_alert_settings(metric, resource='memory')
        self._append_metric_data(
            metric, percent_used, current, time=time, extra_values=extra_values
        )

    def _calculate_increment(self, ifname, stat, value):
        """Returns how much a counter has incremented since its last saved value."""
        # get previous counters
        data = self._previous_data
        try:
            previous_counter = data['interfaces_dict'][ifname]['statistics'][stat]
        except KeyError:
            # if no previous measurements present, counter will start from zero
            previous_counter = 0
        # if current value is higher than previous value,
        # it means the interface traffic counter is increasing
        # and to calculate the traffic performed since the last
        # measurement we have to calculate the difference
        if value >= previous_counter:
            return int(value - previous_counter)
        # on the other side, if the current value is less than
        # the previous value, it means that the counter was restarted
        # (eg: reboot, configuration reload), so we keep the whole amount
        else:
            return int(value)

    def _create_traffic_chart(self, metric):
        """Creates "traffic (GB)" chart."""
        if 'traffic' not in monitoring_settings.AUTO_CHARTS:
            return
        chart = Chart(metric=metric, configuration='traffic')
        chart.full_clean()
        chart.save()

    def _create_clients_chart(self, metric):
        """Creates "WiFi associations" chart."""
        if 'wifi_clients' not in monitoring_settings.AUTO_CHARTS:
            return
        chart = Chart(metric=metric, configuration='wifi_clients')
        chart.full_clean()
        chart.save()

    def _create_resources_chart(self, metric, resource):
        if resource not in monitoring_settings.AUTO_CHARTS:
            return
        chart = Chart(metric=metric, configuration=resource)
        chart.full_clean()
        chart.save()

    def _create_resources_alert_settings(self, metric, resource):
        alert_settings = AlertSettings(metric=metric)
        alert_settings.full_clean()
        alert_settings.save()

    def _create_signal_strength_chart(self, metric):
        chart = Chart(metric=metric, configuration='signal_strength')
        chart.full_clean()
        chart.save()

    def _create_signal_quality_chart(self, metric):
        chart = Chart(metric=metric, configuration='signal_quality')
        chart.full_clean()
        chart.save()

    def _create_access_tech_chart(self, metric):
        chart = Chart(metric=metric, configuration='access_tech')
        chart.full_clean()
        chart.save()



# Generated by Django 3.2.12 on 2022-04-23 17:50

import collections

import jsonfield.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('monitoring', '0003_populate_metric_is_healthy_tolerance'),
    ]

    operations = [
        migrations.AddField(
            model_name='metric',
            name='main_tags',
            field=jsonfield.fields.JSONField(
                blank=True,
                db_index=True,
                default=dict,
                dump_kwargs={'indent': 4},
                load_kwargs={'object_pairs_hook': collections.OrderedDict},
                verbose_name='main tags',
            ),
        ),
        migrations.AddField(
            model_name='metric',
            name='extra_tags',
            field=jsonfield.fields.JSONField(
                blank=True,
                default=dict,
                dump_kwargs={'indent': 4},
                load_kwargs={'object_pairs_hook': collections.OrderedDict},
                verbose_name='extra tags',
            ),
        ),
        migrations.AlterUniqueTogether(
            name='metric',
            unique_together={
                ('key', 'field_name', 'content_type', 'object_id', 'main_tags')
            },
        ),
    ]



# Generated by Django 3.2.14 on 2022-08-12 11:04

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('monitoring', '0008_create_general_metrics'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='alertsettings',
            options={
                'permissions': (
                    ('add_alertsettings_inline', 'Can add Alert settings inline'),
                    ('change_alertsettings_inline', 'Can change Alert settings inline'),
                    ('delete_alertsettings_inline', 'Can delete Alert settings inline'),
                    ('view_alertsettings_inline', 'Can view Alert settings inline'),
                ),
                'verbose_name': 'Alert settings',
                'verbose_name_plural': 'Alert settings',
            },
        ),
    ]
